import math
import os
import pathlib
import shutil
import sys
from functools import lru_cache
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import pandas as pd
import rich
import structlog
from rich.table import Table

import rasa.shared.data
import rasa.shared.utils.cli
import rasa.utils.io
from rasa.e2e_test.constants import (
    CONFTEST_FILENAMES,
    E2E_CONFTEST_SCHEMA_FILE_PATH,
    KEY_FIXTURES,
    KEY_METADATA,
    KEY_STUB_CUSTOM_ACTIONS,
    KEY_TEST_CASE,
    KEY_TEST_CASES,
    STATUS_FAILED,
    STATUS_PASSED,
    STUB_CUSTOM_ACTION_NAME_SEPARATOR,
)
from rasa.e2e_test.e2e_test_case import (
    Fixture,
    Metadata,
    TestCase,
    TestCaseFixtures,
    TestSuite,
)
from rasa.e2e_test.stub_custom_action import (
    StubCustomAction,
    get_stub_custom_action_key,
)
from rasa.e2e_test.utils.fixture_utils import extract_test_case_fixtures
from rasa.e2e_test.utils.validation import (
    read_e2e_test_schema,
    validate_path_to_test_cases,
    validate_test_case,
)
from rasa.exceptions import ValidationError
from rasa.shared.exceptions import DuplicateFixtureException, RasaException
from rasa.shared.utils.yaml import (
    is_key_in_yaml,
    parse_raw_yaml,
    read_schema_file,
    validate_yaml_content_using_schema,
    validate_yaml_data_using_schema_with_assertions,
)
from rasa.utils.beta import BetaNotEnabledException, ensure_beta_feature_is_enabled

if TYPE_CHECKING:
    from rasa.e2e_test.aggregate_test_stats_calculator import AccuracyCalculation
    from rasa.e2e_test.e2e_test_result import TestResult


RASA_PRO_BETA_STUB_CUSTOM_ACTION_ENV_VAR_NAME = "RASA_PRO_BETA_STUB_CUSTOM_ACTION"

structlogger = structlog.get_logger()


def color_difference(diff: List[str]) -> Generator[str, None, None]:
    """Colorize the difference between two strings.

    Example:
        >>> color_difference(["+ Hello", "- World"])
        ["<ansigreen>+ Hello</ansigreen>", "<ansired>- World</ansired>"]

    Args:
        diff: List of lines of the diff.

    Returns:
    Generator of colored lines.
    """
    for line in diff:
        if line.startswith("+"):
            yield "[green3]" + line + "[/green3]"
        elif line.startswith("-"):
            yield "[red3]" + line + "[/red3]"
        elif line.startswith("^"):
            yield "[blue3]" + line + "[/blue3]"
        elif line.startswith("?"):
            yield "[grey37]" + line + "[/grey37]"
        else:
            yield line


def print_failed_case(fail: "TestResult") -> None:
    """Print the details of a failed test case.

    Example:
        >>> print_failed_case(TestResult(TestCase([TestStep()]), True,
        ...                  ["- Hello", "+ World"]))
        ---------------------- test in test.md failed ----------------------
        Mismatch starting at test.md:1:
        <ansired>- Hello</ansired>
        <ansigreen>+ World</ansigreen>
    """
    fail_headline = (
        f"'{fail.test_case.name}' in {fail.test_case.file_with_line()} failed"
    )
    rasa.shared.utils.cli.print_error(
        f"{rasa.shared.utils.cli.pad(fail_headline, char='-')}\n"
    )
    rasa.shared.utils.cli.print_error(
        f"Mismatch starting at {fail.test_case.file}:{fail.error_line}: \n"
    )
    if fail.difference:
        rich.print(("\n".join(color_difference(fail.difference))))

    if fail.assertion_failure:
        rasa.shared.utils.cli.print_error(
            f"Assertion type '{fail.assertion_failure.assertion.type()}' failed "
            f"with this error message: {fail.assertion_failure.error_message}\n"
        )
        rasa.shared.utils.cli.print_error("Actual events transcript:\n")
        rasa.shared.utils.cli.print_error(
            "\n".join(fail.assertion_failure.actual_events_transcript)
        )


def print_test_summary(failed: List["TestResult"]) -> None:
    """Print the summary of the test run.

    Example:
        >>> print_test_summary([TestResult(TestCase([TestStep()]), True,
        ...                  ["- Hello", "+ World"])])
        =================== short test summary info ===================
        FAILED test.md::test
    """
    rasa.shared.utils.cli.print_info(
        rasa.shared.utils.cli.pad("short test summary info")
    )

    for f in failed:
        rasa.shared.utils.cli.print_error(
            f"FAILED {f.test_case.file}::{f.test_case.name}"
        )


def print_final_line(
    passed: List["TestResult"], failed: List["TestResult"], has_failed: bool
) -> None:
    """Print the final line of the test output.

    Args:
        passed: List of passed test cases.
        failed: List of failed test cases.
        has_failed: Boolean, true if the test run has failed.
    """
    final_line_color = "green3" if not has_failed else "red3"

    width = shutil.get_terminal_size((80, 20)).columns

    # calculate the length of the text - this is a bit hacky but works
    text_lengt = (
        math.ceil(len(passed) / 10)  # length of the number of passed tests
        + math.ceil(len(failed) / 10)  # length of the number of failed tests
        + 18  # length of the text "  failed, passed  "
    )
    # we can't use the padding function here as the text contains html tags
    # which are not taken into account when calculating the length
    padding = max(6, width - text_lengt)
    pre_pad = "=" * max(3, padding // 2)
    post_pad = "=" * max(3, math.ceil(padding / 2))
    rich.print(
        f"[{final_line_color}]{pre_pad} "
        f"[bold red3]{len(failed)} failed[/bold red3]"
        f"[bright_white], [/bright_white]"
        f"[bold green3]{len(passed)} passed[/bold green3]"
        f" {post_pad}[/{final_line_color}]"
    )


def print_aggregate_stats(accuracy_calculations: List["AccuracyCalculation"]) -> None:
    """Print the aggregate statistics of the test run."""
    rasa.shared.utils.cli.print_info(
        rasa.shared.utils.cli.pad("Accuracy By Assertion Type")
    )
    table = Table()
    table.add_column("Assertion Type", justify="center", style="cyan")
    table.add_column("Accuracy", justify="center", style="cyan")

    for accuracy_calculation in accuracy_calculations:
        table.add_row(
            accuracy_calculation.assertion_type, f"{accuracy_calculation.accuracy:.2%}"
        )

    rich.print(table)


def print_test_result(
    passed: List["TestResult"],
    failed: List["TestResult"],
    fail_fast: bool = False,
    **kwargs: Any,
) -> None:
    """Print the result of the test run.

    Args:
        passed: List of passed test cases.
        failed: List of failed test cases.
        fail_fast: If true, stop after the first failure.
        **kwargs: additional arguments
    """
    if failed:
        # print failure headline
        print("\n")
        rich.print(f"[bold]{rasa.shared.utils.cli.pad('FAILURES', char='=')}[/bold]")

    # print failed test_Case
    for fail in failed:
        print_failed_case(fail)

    accuracy_calculations = kwargs.get("accuracy_calculations", [])
    if accuracy_calculations:
        print_aggregate_stats(accuracy_calculations)

    print_test_summary(failed)

    if fail_fast:
        rasa.shared.utils.cli.print_error(
            rasa.shared.utils.cli.pad("stopping after 1 failure", char="!")
        )
        has_failed = True
    elif len(failed) + len(passed) == 0:
        # no tests were run, print error
        rasa.shared.utils.cli.print_error(
            rasa.shared.utils.cli.pad("no test cases found", char="!")
        )
        print_e2e_help()
        has_failed = True
    elif failed:
        has_failed = True
    else:
        has_failed = False

    print_final_line(passed, failed, has_failed=has_failed)
    sys.exit(1 if has_failed else 0)


def print_e2e_help() -> None:
    """Print help guiding users how to write e2e tests."""
    rasa.shared.utils.cli.print_info(
        dedent(
            """\
        To start using e2e tests create a yaml file in a test directory, e.g.
        'tests/test_cases.yml'. You can find example test cases in the starter
        pack at

            https://github.com/RasaHQ/starter-pack-intentless-policy#testing-the-policy

        Here is an example of a test case in yaml format:

            test_cases:
            - test_case: "test_greet"
              steps:
              - user: "hello there!"
              - bot: "Hey! How are you?"

        To run the e2e tests, execute:
            >>> rasa test e2e <path-to-test-cases>
    """
        )
    )


def split_into_passed_failed(
    results: List["TestResult"],
) -> Tuple[List["TestResult"], List["TestResult"]]:
    """Get the summary of the test results.

    Args:
        results: List of test results.

    Returns:
        Tuple consisting of passed count, failed count and failed test cases.
    """
    passed_cases = [r for r in results if r.pass_status]
    failed_cases = [r for r in results if not r.pass_status]

    return passed_cases, failed_cases


@lru_cache(maxsize=1)
def extract_test_case_from_path(path: str) -> Tuple[str, str]:
    """Extract test case from path if specified.

    Args:
        path: Path to the file or folder containing test cases.

    Returns:
        Tuple consisting of the path to test cases and the extracted test case name.
    """
    test_case_name = ""

    if "::" in str(path):
        splitted_path = path.split("::")
        test_case_name = splitted_path[-1]
        path = splitted_path[0]

    return path, test_case_name


def is_test_case_file(file_path: Union[str, Path]) -> bool:
    """Check if file contains test cases.

    Args:
        file_path: Path of the file to check.

    Returns:
        `True` if the file contains test cases, `False` otherwise.
    """
    return rasa.shared.data.is_likely_yaml_file(file_path) and is_key_in_yaml(
        file_path, KEY_TEST_CASES
    )


def get_conftest_files_for_directory(dir_path: str) -> List[str]:
    """List conftest files from directory up to root (root first, then leaf).

    Args:
        dir_path: Absolute or relative directory path.

    Returns:
        Ordered list of conftest file paths: root conftest first, then child, etc.
    """
    dir_path = os.path.abspath(os.path.normpath(dir_path))
    if not os.path.isdir(dir_path):
        dir_path = os.path.dirname(dir_path)
    chain: List[str] = []
    current: Optional[str] = dir_path
    while current:
        for name in CONFTEST_FILENAMES:
            p = os.path.join(current, name)
            if os.path.isfile(p):
                chain.append(p)
                break
        current = (
            os.path.dirname(current) if current != os.path.dirname(current) else None
        )
    chain.reverse()
    return chain


def extract_test_cases(
    test_file_content: dict,
    test_case_name: str,
    test_file: str,
) -> List[TestCase]:
    """Extract test cases from the test file content.

    Args:
        test_file_content: Content of the test file.
        test_case_name: Name of the test case to extract.
        test_file: Path to the test file.

    Returns:
        List of test cases.
    """
    test_cases_content = test_file_content.get(KEY_TEST_CASES) or []
    if test_case_name:
        return [
            TestCase.from_dict(test_case_dict, file=test_file)
            for test_case_dict in test_cases_content
            if test_case_name == test_case_dict.get(KEY_TEST_CASE)
        ]
    return [
        TestCase.from_dict(test_case_dict, file=test_file)
        for test_case_dict in test_cases_content
    ]


def extract_fixtures_from_content(
    content: dict,
    source_file: Optional[str] = None,
) -> Dict[str, Fixture]:
    """Extract fixtures from YAML content. Only within-file duplicates raise.

    Used for conftest and per-file chain building where later files override
    earlier; cross-file duplicates are allowed.

    Args:
        content: Parsed YAML dict (must have KEY_FIXTURES).
        source_file: Path for error messages.

    Returns:
        Dict of fixture name -> Fixture (no duplicates within this content).

    Raises:
        DuplicateFixtureException: If the same fixture name appears twice.
    """
    fixtures_content = content.get(KEY_FIXTURES) or []
    seen: set = set()
    duplicates: List[str] = []
    fixture_objs: List[Fixture] = []

    for fixture in fixtures_content:
        fixture_obj = Fixture.from_dict(fixture_dict=fixture)
        if fixture_obj.name in seen:
            duplicates.append(fixture_obj.name)
            continue
        seen.add(fixture_obj.name)
        fixture_objs.append(fixture_obj)

    if duplicates:
        raise DuplicateFixtureException(duplicates, source_file or "")

    return {f.name: f for f in fixture_objs}


def extract_metadata(
    test_file_content: dict, existing_metadata: Dict[str, Metadata]
) -> Dict[str, Metadata]:
    """Extract metadata from the test file content.

    Args:
        test_file_content: Content of the test file.
        existing_metadata: Existing metadata.

    Returns:
        Dict of metadata.
    """
    metadata_contents = test_file_content.get(KEY_METADATA) or []
    _metadata = {}
    for metadata_content in metadata_contents:
        metadata_obj = Metadata.from_dict(metadata_dict=metadata_content)
        if existing_metadata.get(metadata_obj.name) is None:
            _metadata[metadata_obj.name] = metadata_obj
    return _metadata


def extract_stub_custom_actions(
    test_file_content: dict, test_file: str
) -> Dict[str, StubCustomAction]:
    """Extract stub custom actions from the test file content.

    Args:
        test_file_content: Content of the test file.
        test_file: Path to the test file.

    Returns:
        Dict of stub custom actions.
    """
    stub_custom_actions_contents = test_file_content.get(KEY_STUB_CUSTOM_ACTIONS) or {}
    _stub_custom_actions = {}
    for action_name, stub_data in stub_custom_actions_contents.items():
        if STUB_CUSTOM_ACTION_NAME_SEPARATOR in action_name:
            stub_custom_action_key = action_name
        else:
            test_file_name = Path(test_file).name
            stub_custom_action_key = get_stub_custom_action_key(
                test_file_name, action_name
            )
        _stub_custom_actions[stub_custom_action_key] = StubCustomAction.from_dict(
            action_name=action_name, stub_data=stub_data
        )
    return _stub_custom_actions


def _load_conftest(
    conftest_path: str,
    conftest_schema: Union[Dict[str, Any], List[Any]],
) -> Dict[str, Fixture]:
    """Load and parse a single conftest file (fixtures only).

    Uses conftest schema (fixtures only) so files that also contain
    llm_judge or other config keys pass validation.
    """
    content = parse_raw_yaml(Path(conftest_path).read_text(encoding="utf-8"))
    validate_yaml_content_using_schema(content, conftest_schema)
    return extract_fixtures_from_content(content, conftest_path)


def _merge_fixture_chain(chain: List[Dict[str, Fixture]]) -> Dict[str, Fixture]:
    """Merge fixture dicts in order; later overrides earlier. O(n) single pass."""
    merged: Dict[str, Fixture] = {}
    for d in chain:
        merged.update(d)
    return merged


def _load_and_validate_test_file(
    test_file: str,
    e2e_test_schema: Union[Dict[str, Any], List[Any]],
) -> Dict[str, Any]:
    """Parse and schema-validate a single test file. Returns parsed YAML content."""
    test_file_content = parse_raw_yaml(Path(test_file).read_text(encoding="utf-8"))
    validate_yaml_data_using_schema_with_assertions(
        yaml_data=test_file_content, schema_content=e2e_test_schema
    )
    return test_file_content


def _get_resolved_fixtures_for_test_file(
    test_file: str,
    test_file_content: Dict[str, Any],
    conftest_cache: Dict[str, Dict[str, Fixture]],
    conftest_schema: Union[Dict[str, Any], List[Any]],
) -> Tuple[Optional[Dict[str, Fixture]], Optional[str]]:
    """Resolve fixtures for a test file: parent conftests first, then child overrides.

    Fetches all parent conftest files (root first), then adds the test file's own
    fixtures. Later definitions override earlier (child overrides parent).

    Returns:
        (merged_fixtures_dict, None) on success, or (None, error_message) on
        DuplicateFixtureException or ValidationError.
    """
    chain: List[Dict[str, Fixture]] = []
    for conftest_path in get_conftest_files_for_directory(os.path.dirname(test_file)):
        if conftest_path not in conftest_cache:
            try:
                conftest_cache[conftest_path] = _load_conftest(
                    conftest_path, conftest_schema
                )
            except (DuplicateFixtureException, ValidationError) as e:
                return (None, str(e))
        if conftest_path in conftest_cache:
            chain.append(conftest_cache[conftest_path])

    try:
        chain.append(extract_fixtures_from_content(test_file_content, test_file))
    except (DuplicateFixtureException, ValidationError) as e:
        return (None, str(e))

    merged = _merge_fixture_chain(chain)
    return (merged, None)


def read_test_cases(path: str) -> TestSuite:
    """Read test cases from the given path.

    For every test file:
    1. Load and validate the test file and get test file content.
    2. Extract test cases from the test file.
    3. Extract fixtures for the test file by fetching all parent conftest files
       and overriding with child (test file) fixtures.
    4. Create an entry in fixtures_per_test for each test case with test_name,
       file_name, and associated fixtures.

    Conftest files (conftest.yml / conftest.yaml) are loaded from the path's
    directory and parent directories; when the same fixture name is defined in
    a conftest and in a test file, the test file's definition wins for that file.

    Args:
        path: Path to the file or folder containing test cases.

    Returns:
        TestSuite.
    """
    path, test_case_name = extract_test_case_from_path(path)
    validate_path_to_test_cases(path)

    test_files = rasa.shared.data.get_data_files([path], is_test_case_file)
    # read the e2e test schema
    e2e_test_schema = read_e2e_test_schema()
    # read the conftest schema
    conftest_schema = read_schema_file(E2E_CONFTEST_SCHEMA_FILE_PATH)

    input_test_cases: List[TestCase] = []
    metadata: Dict[str, Metadata] = {}
    stub_custom_actions: Dict[str, StubCustomAction] = {}
    fixtures_per_test: List[TestCaseFixtures] = []
    fixture_errors: List[str] = []
    conftest_cache: Dict[str, Dict[str, Fixture]] = {}

    for test_file in test_files:
        # 1. Load and validate the test file and get test file content
        test_file_content = _load_and_validate_test_file(test_file, e2e_test_schema)

        # 2. Extract test cases from the test file
        test_cases = extract_test_cases(test_file_content, test_case_name, test_file)

        # 3. Extract resolved fixtures for the test file.
        resolved_fixtures, fixture_error = _get_resolved_fixtures_for_test_file(
            test_file, test_file_content, conftest_cache, conftest_schema
        )
        if fixture_error:
            fixture_errors.append(fixture_error)
            continue

        # 4. Create an entry in fixtures_per_test for each test case
        # (only fixtures used by that test)
        if resolved_fixtures is not None:
            fixtures_per_test.extend(
                extract_test_case_fixtures(test_cases, list(resolved_fixtures.values()))
            )
        input_test_cases.extend(test_cases)
        metadata.update(extract_metadata(test_file_content, metadata))
        stub_custom_actions.update(
            extract_stub_custom_actions(test_file_content, test_file)
        )

    if fixture_errors:
        raise RasaException(
            "Errors while loading fixtures from file(s):\n  - "
            + "\n  - ".join(fixture_errors)
        )

    validate_test_case(test_case_name, input_test_cases, fixtures_per_test, metadata)
    if stub_custom_actions:
        check_beta_feature_flag_for_custom_actions_stubs()

    return TestSuite(
        test_cases=input_test_cases,
        fixtures_per_test=fixtures_per_test,
        metadata=list(metadata.values()),
        stub_custom_actions=stub_custom_actions,
    )


def check_beta_feature_flag_for_custom_actions_stubs() -> None:
    """Check the beta feature flag for custom actions stub and exit if not enabled."""
    try:
        ensure_beta_feature_is_enabled(
            "enabling stubs for custom actions",
            RASA_PRO_BETA_STUB_CUSTOM_ACTION_ENV_VAR_NAME,
        )
    except BetaNotEnabledException as exc:
        rasa.shared.utils.cli.print_error_and_exit(str(exc))


def _save_coverage_report(
    report: Optional[pd.DataFrame], test_status: str, output_dir: str
) -> None:
    if report is None:
        return

    if report is not None and not report.empty:
        if test_status == STATUS_PASSED:
            rasa.shared.utils.cli.print_success(report.to_string(index=False))
        else:
            rasa.shared.utils.cli.print_error(report.to_string(index=False))

    output_filename = f"coverage_report_for_{test_status}_tests.csv"
    output_file_path = os.path.join(output_dir, output_filename)
    report.to_csv(output_file_path, index=False)
    structlogger.info(
        "rasa.e2e_test.save_coverage_report",
        message=f"Coverage result for {test_status} e2e tests"
        f" is written to '{output_file_path}'.",
    )


def write_test_results_to_file(results: List["TestResult"], output_file: str) -> None:
    """Write test results to a file.

    Args:
        results: List of test results.
        output_file: Path to the output file.
    """
    Path(output_file).touch()

    data = {"test_results": [test_result.as_dict() for test_result in results]}

    rasa.utils.io.write_yaml(
        data, target=output_file, transform=transform_results_output_to_yaml
    )

    if STATUS_PASSED in output_file:
        rasa.shared.utils.cli.print_info(
            f"Passing test results have been saved at path: {output_file}."
        )
    elif STATUS_FAILED in output_file:
        rasa.shared.utils.cli.print_info(
            f"Failing test results have been saved at path: {output_file}."
        )


def transform_results_output_to_yaml(yaml_string: str) -> str:
    """Transform the output of the YAML writer to make it more readable.

    Args:
        yaml_string: The YAML string to transform.

    Returns:
        The transformed YAML string.
    """
    result = []
    for s in yaml_string.splitlines(True):
        if s.startswith("- name"):
            result.append("\n")
            result.append(s)
        elif s.startswith("\n"):
            result.append(s.strip())
        elif s.strip().startswith("#"):
            continue
        else:
            result.append(s)
    return "".join(result)


def _save_tested_commands_histogram(
    count_dict: Dict[str, int], test_status: str, output_dir: str
) -> None:
    """Creates a command histogram and saves it to the specified directory.

    Args:
        count_dict (Dict[str, int]): A dictionary where keys are commands
        and values are counts.
        test_status (str): passing or failing
        output_dir (str): The directory path where the histogram
        image will be saved.
    """
    if not count_dict:
        return

    # Sort the dictionary by keys
    sorted_count_dict = dict(sorted(count_dict.items()))

    plt.figure(figsize=(10, 6))
    bars = plt.bar(sorted_count_dict.keys(), sorted_count_dict.values(), color="blue")
    plt.xlabel("Commands")
    plt.ylabel("Counts")
    plt.title(f"Command histogram for {test_status} tests")
    plt.xticks(rotation=45, ha="right")
    plt.tight_layout()

    # Add total number to each bar
    for bar in bars:
        yval = bar.get_height()
        plt.text(
            bar.get_x() + bar.get_width() / 2,
            yval + 0.5,
            int(yval),
            ha="center",
            va="bottom",
        )

    output_filename = f"commands_histogram_for_{test_status}_tests.png"
    output_file_path = pathlib.Path().joinpath(output_dir, output_filename)
    plt.savefig(output_file_path)
    plt.close()

    structlogger.info(
        "rasa.e2e_test._save_tested_commands_histogram",
        message=f"Commands histogram for {test_status} e2e tests "
        f"are written to '{output_file_path}'.",
    )


def save_test_cases_to_yaml(
    test_results: List["TestResult"],
    output_dir: str,
    status: str,
    test_suite: TestSuite,
) -> None:
    """Extracts TestCases from a list of TestResults and saves them to a YAML file."""
    if not test_results:
        return

    test_cases = [result.test_case for result in test_results]
    new_test_suite = TestSuite(
        test_cases=test_cases,
        fixtures_per_test=test_suite.fixtures_per_test,
        metadata=test_suite.metadata,
        stub_custom_actions=test_suite.stub_custom_actions,
    )

    output_filename = f"{status}.yml"
    output_file_path = os.path.join(output_dir, output_filename)
    rasa.utils.io.write_yaml(new_test_suite.as_writable(), target=output_file_path)

    structlogger.info(
        "rasa.e2e_test.save_e2e_test_cases",
        message=f"E2e tests with '{status}' status are written to file: "
        f"'{output_file_path}'.",
    )
    if status == STATUS_PASSED:
        structlogger.info(
            "rasa.e2e_test.save_e2e_test_cases",
            message=f"You can use the file: '{output_file_path}' in case you want to "
            f"create training data for fine-tuning an LLM via "
            f"'rasa llm finetune prepare-data'.",
        )



from typing import Any, Dict, Optional

import structlog
from pydantic import BaseModel, Field

structlogger = structlog.get_logger()


class Document(BaseModel):
    """Model for document retrieval results."""

    content: str = Field(...)
    url: Optional[str] = Field(default=None)
    title: Optional[str] = Field(default=None)
    metadata: Optional[Dict[str, Any]] = Field(default=None)

    @classmethod
    def from_inkeep_rag_response(cls, rag_item: Dict[str, Any]) -> "Document":
        """Create a Document object from a single InKeep RAG response item.

        Args:
            rag_item: Single item from InKeep RAG response

        Returns:
            Document object with extracted content and metadata
        """
        source = rag_item.get("source", {})
        text_content = cls._extract_text_from_source(source)

        return cls(
            content=text_content.strip() if text_content else "",
            url=rag_item.get("url"),
            title=rag_item.get("title"),
            metadata={
                "type": rag_item.get("type"),
                "record_type": rag_item.get("record_type"),
                "context": rag_item.get("context"),
                "media_type": source.get("media_type"),
            },
        )

    @staticmethod
    def _extract_text_from_source(source: Dict[str, Any]) -> str:
        """Extract text content from InKeep source object.

        Args:
            source: Source object from InKeep RAG response

        Returns:
            Extracted text content
        """
        # Try to extract from content array first
        if "content" in source:
            text_parts = []
            for content_item in source["content"]:
                if content_item.get("type") == "text" and content_item.get("text"):
                    text_parts.append(content_item["text"])
            if text_parts:
                return "\n".join(text_parts)

        # Fallback to source data
        return source.get("data", "")



import argparse
from typing import List, Text

from rasa.cli import SubParsersAction
from rasa.shared.constants import (
    DEFAULT_DOMAIN_PATH,
)
from rasa.studio.link import handle_link


def add_subparser(
    subparsers: SubParsersAction,
    parents: List[argparse.ArgumentParser],
    domain: Text = DEFAULT_DOMAIN_PATH,
) -> None:
    """Register the `rasa studio link` command with the main CLI.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
        domain: Path to the assistant's domain file.
    """
    link_parser = subparsers.add_parser(
        "link",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Link the current project to an existing Rasa Studio assistant.",
    )

    link_parser.add_argument(
        "assistant_name",
        type=str,
        help="Name of the assistant in Rasa Studio.",
    )
    link_parser.set_defaults(func=handle_link)



"""Utilities for parsing FAQ-style documents (Q/A pairs) used in extractive search."""

import re
from collections import defaultdict
from typing import TYPE_CHECKING, List

import structlog

from rasa.shared.constants import (
    DOCUMENT_TYPE_FAQ,
    FAQ_DOCUMENT_ENTRY_SEPARATOR,
    FAQ_DOCUMENT_LINE_SEPARATOR,
    FAQ_DOCUMENT_METADATA_ANSWER,
    FAQ_DOCUMENT_METADATA_TITLE,
    FAQ_DOCUMENT_METADATA_TYPE,
    FAQ_INPUT_DATA_ANSWER_LINE_PREFIX,
    FAQ_INPUT_DATA_QUESTION_LINE_PREFIX,
)

if TYPE_CHECKING:
    from langchain.schema import Document

_FAQ_PAIR_PATTERN = re.compile(
    rf"{re.escape(FAQ_INPUT_DATA_QUESTION_LINE_PREFIX)}\s*"
    rf"(?P<question>.*?)\s*{FAQ_DOCUMENT_LINE_SEPARATOR}\s*"
    rf"{re.escape(FAQ_INPUT_DATA_ANSWER_LINE_PREFIX)}\s*"
    rf"(?P<answer>.*)",
    re.DOTALL,
)


structlogger = structlog.get_logger()


def _format_faq_documents(documents: List["Document"]) -> List["Document"]:
    """Splits each loaded file into individual FAQs.

    Args:
        documents: Documents representing whole files containing FAQs.

    Returns:
        List of Document objects, each containing a separate FAQ.

    Examples:
        An example of a file containing FAQs:

        Q: Who is Finley?
        A: Finley is your smart assistant for the FinX App. You can add him to your
           favorite messenger and tell him what you need help with.

        Q: How does Finley work?
        A: Finley is powered by the latest chatbot technology leveraging a unique
           interplay of large language models and secure logic.

    More details in documentation: https://rasa.com/docs/reference/config/policies/extractive-search/
    """
    structured_faqs = []
    from langchain.schema import Document

    for document in documents:
        chunks = document.page_content.strip().split(FAQ_DOCUMENT_ENTRY_SEPARATOR)

        for chunk in chunks:
            match = _FAQ_PAIR_PATTERN.match(chunk.strip())

            if not match:
                structlogger.warning(
                    "faq_parser.format_faq_documents.invalid_chunk_skipped",
                    event_info=(
                        "Chunk does not match expected QA format. "
                        "Please refer to the documentation: "
                        "https://rasa.com/docs/reference/config/"
                        "policies/extractive-search/"
                    ),
                    chunk_preview=chunk[:100],
                )
                continue

            question = match.group("question").strip()
            answer = match.group("answer").strip()
            title = _sanitize_title(question)

            formatted_document = Document(
                page_content=question,
                metadata={
                    FAQ_DOCUMENT_METADATA_TITLE: title,
                    FAQ_DOCUMENT_METADATA_TYPE: DOCUMENT_TYPE_FAQ,
                    FAQ_DOCUMENT_METADATA_ANSWER: answer,
                },
            )

            structured_faqs.append(formatted_document)

            structlogger.debug(
                "faq_parser.format_faq_documents.parsed_chunk",
                event_info="Parsed chunk.",
                title=title,
                question=question,
                answer=answer,
                parsed_chunk_preview=chunk[:100],
            )

    structlogger.debug(
        "faq_parser.format_faq_documents.parsed_chunks",
        event_info=(
            f"Retrieved {len(structured_faqs)} FAQ pair(s)"
            f"from {len(documents)} document(s)."
        ),
        num_structured_faqs=len(structured_faqs),
        num_documents=len(documents),
    )
    _check_and_parsed_faq_documents_for_duplicates(structured_faqs)
    return structured_faqs


def _sanitize_title(title: str) -> str:
    title = title.lower()
    # Remove all whitespaces with "_"
    title = re.sub(r"\s+", "_", title)
    # Remove all non alpha-numeric characters
    title = re.sub(r"[^\w]", "", title)
    # Collapse multiple "_"
    title = re.sub(r"_+", "_", title)
    # Clean up edges
    return title.strip("_")


def _check_and_parsed_faq_documents_for_duplicates(documents: List["Document"]) -> None:
    seen_qa_pairs = set()
    seen_questions: defaultdict = defaultdict(list)

    for doc in documents:
        question = doc.page_content.strip()
        answer = doc.metadata.get(FAQ_DOCUMENT_METADATA_ANSWER, "").strip()

        if not question or not answer:
            continue

        if (question, answer) in seen_qa_pairs:
            structlogger.warning(
                "faq_parser.duplicate_qa_pair_found",
                event_info="Duplicate QA pair found.",
                question=question,
                answer_preview=answer,
            )
            continue

        if question in seen_questions and seen_questions[question] != answer:
            structlogger.warning(
                "faq_parser.inconsistent_answer",
                event_info="Duplicate question with different answer found.",
                question=question,
                previous_answers=seen_questions[question],
                new_answer=answer,
            )

        seen_qa_pairs.add((question, answer))
        seen_questions[question].append(answer)



import itertools
import logging
import os
from functools import wraps
from typing import Any, Callable, List, Optional, Text, Tuple, TypeVar, Union

import matplotlib
import numpy as np
from matplotlib.ticker import FormatStrFormatter

import rasa.shared.utils.io
from rasa.constants import RESULTS_FILE

logger = logging.getLogger(__name__)


def _fix_matplotlib_backend() -> None:
    """Tries to fix a broken matplotlib backend."""
    try:
        backend = matplotlib.get_backend()
    except Exception:  # skipcq:PYL-W0703
        logger.error(
            "Cannot retrieve Matplotlib backend, likely due to a compatibility "
            "issue with system dependencies. Please refer to the documentation: "
            "https://matplotlib.org/stable/tutorials/introductory/usage.html#backends"
        )
        raise

    # At first, matplotlib will be initialized with default OS-specific
    # available backend
    if backend == "TkAgg":
        try:
            # on OSX sometimes the tkinter package is broken and can't be imported.
            # we'll try to import it and if it fails we will use a different backend
            import tkinter
        except (ImportError, ModuleNotFoundError):
            logger.debug("Setting matplotlib backend to 'agg'")
            matplotlib.use("agg")

    # if no backend is set by default, we'll try to set it up manually
    elif backend is None:  # pragma: no cover
        try:
            # If the `tkinter` package is available, we can use the `TkAgg` backend
            import tkinter  # noqa: F401

            logger.debug("Setting matplotlib backend to 'TkAgg'")
            matplotlib.use("TkAgg")
        except (ImportError, ModuleNotFoundError):
            logger.debug("Setting matplotlib backend to 'agg'")
            matplotlib.use("agg")


ReturnType = TypeVar("ReturnType")
FuncType = Callable[..., ReturnType]
_MATPLOTLIB_BACKEND_FIXED = False


def _needs_matplotlib_backend(func: FuncType) -> FuncType:
    """Decorator to fix matplotlib backend before calling a function."""

    @wraps(func)
    def inner(*args: Any, **kwargs: Any) -> ReturnType:  # type: ignore
        """Replacement function that fixes matplotlib backend."""
        global _MATPLOTLIB_BACKEND_FIXED
        if not _MATPLOTLIB_BACKEND_FIXED:
            _fix_matplotlib_backend()
            _MATPLOTLIB_BACKEND_FIXED = True
        return func(*args, **kwargs)

    return inner


@_needs_matplotlib_backend
def plot_confusion_matrix(
    confusion_matrix: np.ndarray,
    classes: Union[np.ndarray, List[Text]],
    normalize: bool = False,
    title: Text = "Confusion matrix",
    color_map: Any = None,
    zmin: int = 1,
    output_file: Optional[Text] = None,
) -> None:
    """Print and plot the provided confusion matrix.
    Normalization can be applied by setting `normalize=True`.

    Args:
        confusion_matrix: confusion matrix to plot
        classes: class labels
        normalize: If set to true, normalization will be applied.
        title: title of the plot
        color_map: color mapping
        zmin:
        output_file: output file to save plot to

    """
    import matplotlib.pyplot as plt
    from matplotlib.colors import LogNorm

    zmax = confusion_matrix.max() if len(confusion_matrix) > 0 else 1
    plt.clf()
    if not color_map:
        color_map = plt.cm.get_cmap("Blues")
    plt.imshow(
        confusion_matrix,
        interpolation="nearest",
        cmap=color_map,
        aspect="auto",
        norm=LogNorm(vmin=zmin, vmax=zmax),
    )
    plt.title(title)
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90)
    plt.yticks(tick_marks, classes)

    if normalize:
        confusion_matrix = (
            confusion_matrix.astype("float")
            / confusion_matrix.sum(axis=1)[:, np.newaxis]
        )
        logger.info(f"Normalized confusion matrix: \n{confusion_matrix}")
    else:
        logger.info(f"Confusion matrix, without normalization: \n{confusion_matrix}")

    thresh = zmax / 2.0
    for i, j in itertools.product(
        range(confusion_matrix.shape[0]), range(confusion_matrix.shape[1])
    ):
        plt.text(
            j,
            i,
            confusion_matrix[i, j],
            horizontalalignment="center",
            color="white" if confusion_matrix[i, j] > thresh else "black",
        )

    plt.ylabel("True label")
    plt.xlabel("Predicted label")

    # save confusion matrix to file before showing it
    if output_file:
        fig = plt.gcf()
        fig.set_size_inches(20, 20)
        fig.savefig(output_file, bbox_inches="tight")


def _extract_paired_histogram_specification(
    histogram_data: List[List[float]],
    num_bins: int,
    density: bool,
    x_pad_fraction: float,
    y_pad_fraction: float,
) -> Tuple[List[float], List[List[float]], List[float], Tuple[float, float]]:
    """Extracts all information from the data needed to plot a paired histogram.

    Args:
        histogram_data: Two data vectors
        num_bins: Number of bins to be used for the histogram
        density: If true, generate information for a probability density histogram
        x_pad_fraction: Percentage of extra space in the horizontal direction
        y_pad_fraction: Percentage of extra space in the vertical direction

    Returns:
        The bins, values, ranges of either x-axis, and the range of the y-axis

    Raises:
        ValueError: If histogram_data does not contain values.
    """
    if not histogram_data or not np.concatenate(histogram_data).size:
        rasa.shared.utils.io.raise_warning("No data to plot paired histogram.")
        raise ValueError("No data to plot paired histogram.")
    min_data_value: float = np.min(np.concatenate(histogram_data))
    max_data_value: float = np.max(np.concatenate(histogram_data))
    bin_width = (max_data_value - min_data_value) / num_bins
    bins = [
        min_data_value + i * bin_width
        # `bins` describes the _boundaries_ of the bins, so we need
        # 2 extra - one at the beginning and one at the end
        for i in range(num_bins + 2)
    ]
    histograms = [
        # A list of counts - how often a value in `data` falls into a particular bin
        list(np.histogram(data, bins=bins, density=density)[0])
        for data in histogram_data
    ]

    y_padding = 0.5 * bin_width + y_pad_fraction * bin_width

    if density:
        # Get the maximum count across both histograms, and scale it
        # with `x_pad_fraction`
        v = max([(1.0 + x_pad_fraction) * max(histogram) for histogram in histograms])
        # When we plot the PDF, let both x-axes run to the same value
        # so it's easier to compare visually
        x_ranges = [v, v]
    else:
        # For the left and right histograms, get the largest counts and scale them
        # by `x_pad_fraction` to get the maximum x-values displayed
        x_ranges = [(1.0 + x_pad_fraction) * max(histogram) for histogram in histograms]

    try:
        bin_of_first_non_zero_tally = min(
            [[bool(v) for v in histogram].index(True) for histogram in histograms]
        )
    except ValueError:
        bin_of_first_non_zero_tally = 0

    y_range = (
        # Start plotting where the data starts (ignore empty bins at the low end)
        bins[bin_of_first_non_zero_tally] - y_padding,
        # The y_padding adds half a bin width, as we want the bars to be
        # _centered_ on the bins. We take the next-to-last element of `bins`,
        # because that is the beginning of the last bin.
        bins[-2] + y_padding,
    )

    return bins, histograms, x_ranges, y_range


@_needs_matplotlib_backend
def plot_paired_histogram(
    histogram_data: List[List[float]],
    title: Text,
    output_file: Optional[Text] = None,
    num_bins: int = 25,
    colors: Tuple[Text, Text] = ("#009292", "#920000"),  # (dark cyan, dark red)
    axes_label: Tuple[Text, Text] = ("Correct", "Wrong"),
    frame_label: Tuple[Text, Text] = ("Number of Samples", "Confidence"),
    density: bool = False,
    x_pad_fraction: float = 0.05,
    y_pad_fraction: float = 0.10,
) -> None:
    """Plots a side-by-side comparative histogram of the confidence distribution.

    Args:
        histogram_data: Two data vectors
        title: Title to be displayed above the plot
        output_file: File to save the plot to
        num_bins: Number of bins to be used for the histogram
        colors: Left and right bar colors as hex color strings
        axes_label: Labels shown above the left and right histogram,
            respectively
        frame_label: Labels shown below and on the left of the
            histogram, respectively
        density: If true, generate a probability density histogram
        x_pad_fraction: Percentage of extra space in the horizontal direction
        y_pad_fraction: Percentage of extra space in the vertical direction
    """
    if num_bins <= 2:
        rasa.shared.utils.io.raise_warning(
            f"Number {num_bins} of paired histogram bins must be at least 3."
        )
        return

    try:
        bins, tallies, x_ranges, y_range = _extract_paired_histogram_specification(
            histogram_data,
            num_bins,
            density=density,
            x_pad_fraction=x_pad_fraction,
            y_pad_fraction=y_pad_fraction,
        )
    except (ValueError, TypeError) as e:
        rasa.shared.utils.io.raise_warning(
            f"Unable to plot paired histogram '{title}': {e}"
        )
        return
    yticks = [float(f"{x:.2f}") for x in bins]

    import matplotlib.pyplot as plt

    plt.gcf().clear()

    fig, axes = plt.subplots(ncols=2, sharey=True)
    for side in range(2):
        axes[side].barh(
            bins[:-1],
            tallies[side],
            height=np.diff(bins),
            align="center",
            color=colors[side],
            linewidth=1,
            edgecolor="white",
        )
        axes[side].set(title=axes_label[side])
        axes[side].set(yticks=yticks, xlim=(0, x_ranges[side]), ylim=y_range)

    axes[0].yaxis.set_major_formatter(FormatStrFormatter("%.2f"))
    axes[0].yaxis.set_minor_formatter(FormatStrFormatter("%.2f"))

    axes[0].invert_xaxis()
    axes[0].yaxis.tick_right()

    # Add the title
    fig.suptitle(title, fontsize="x-large", fontweight="bold")

    # Add hidden plot to correctly add x and y labels (frame_label)
    fig.add_subplot(111, frameon=False)

    # Hide tick and tick label of the unused axis
    plt.tick_params(labelcolor="none", top=False, bottom=False, left=False, right=False)
    plt.xlabel(frame_label[0])
    plt.ylabel(frame_label[1])

    if output_file:
        fig = plt.gcf()
        fig.set_size_inches(10, 10)
        fig.tight_layout(w_pad=0)
        fig.savefig(output_file, bbox_inches="tight")


@_needs_matplotlib_backend
def plot_curve(
    output_directory: Text,
    number_of_examples: List[int],
    x_label_text: Text,
    y_label_text: Text,
    graph_path: Text,
) -> None:
    """Plot the results from a model comparison.

    Args:
        output_directory: Output directory to save resulting plots to
        number_of_examples: Number of examples per run
        x_label_text: text for the x axis
        y_label_text: text for the y axis
        graph_path: output path of the plot
    """
    import matplotlib.pyplot as plt

    plt.gcf().clear()

    ax = plt.gca()

    # load results from file
    data = rasa.shared.utils.io.read_json_file(
        os.path.join(output_directory, RESULTS_FILE)
    )
    x = number_of_examples

    # compute mean of all the runs for different configs
    for label in data.keys():
        if len(data[label]) == 0:
            continue
        mean = np.mean(data[label], axis=0)
        std = np.std(data[label], axis=0)
        ax.plot(x, mean, label=label, marker=".")
        ax.fill_between(
            x,
            [m - s for m, s in zip(mean, std)],
            [m + s for m, s in zip(mean, std)],
            color="#6b2def",
            alpha=0.2,
        )
    ax.legend(loc=4)

    ax.set_xlabel(x_label_text)
    ax.set_ylabel(y_label_text)

    plt.savefig(graph_path, format="pdf")

    logger.info(f"Comparison graph saved to '{graph_path}'.")



from typing import List, Text, Tuple

import numpy as np


def cleanup_tokens(
    token_ids_string: List[Tuple[int, Text]], delimiter: Text
) -> Tuple[List[int], List[Text]]:
    """Utility method to apply delimiter based cleanup on list of tokens.

    Args:
        token_ids_string: List of tuples with each tuple containing
            (token id, token string).
        delimiter: character/string to be cleaned from token strings.

    Returns:
        Token ids and Token strings unpacked.
    """
    token_ids_string = [
        (id, string.replace(delimiter, "")) for id, string in token_ids_string
    ]

    # remove empty strings
    token_ids_string = [(id, string) for id, string in token_ids_string if string]

    # return as individual token ids and token strings
    token_ids, token_strings = zip(*token_ids_string)
    # FIXME: zip official typing is not really properly set up
    return token_ids, token_strings  # type: ignore[return-value]


def bert_tokens_pre_processor(token_ids: List[int]) -> List[int]:
    """Add BERT style special tokens(CLS and SEP).

    Args:
        token_ids: List of token ids without any special tokens.

    Returns:
        List of token ids augmented with special tokens.
    """
    BERT_CLS_ID = 101
    BERT_SEP_ID = 102

    processed_tokens = token_ids

    processed_tokens.insert(0, BERT_CLS_ID)
    processed_tokens.append(BERT_SEP_ID)

    return processed_tokens


def gpt_tokens_pre_processor(token_ids: List[int]) -> List[int]:
    """Add GPT style special tokens(None).

    Args:
        token_ids: List of token ids without any special tokens.

    Returns:
        List of token ids augmented with special tokens.
    """
    return token_ids


def xlnet_tokens_pre_processor(token_ids: List[int]) -> List[int]:
    """Add XLNET style special tokens.

    Args:
        token_ids: List of token ids without any special tokens.

    Returns:
        List of token ids augmented with special tokens.
    """
    XLNET_CLS_ID = 3
    XLNET_SEP_ID = 4

    token_ids.append(XLNET_SEP_ID)
    token_ids.append(XLNET_CLS_ID)

    return token_ids


def roberta_tokens_pre_processor(token_ids: List[int]) -> List[int]:
    """Add RoBERTa style special tokens.

    Args:
        token_ids: List of token ids without any special tokens.

    Returns:
        List of token ids augmented with special tokens.
    """
    ROBERTA_BEG_ID = 0
    ROBERTA_END_ID = 2

    token_ids.insert(0, ROBERTA_BEG_ID)
    token_ids.append(ROBERTA_END_ID)

    return token_ids


def xlm_tokens_pre_processor(token_ids: List[int]) -> List[int]:
    """Add XLM style special tokens.

    Args:
        token_ids: List of token ids without any special tokens.

    Returns:
        List of token ids augmented with special tokens.
    """
    XLM_SEP_ID = 1

    token_ids.insert(0, XLM_SEP_ID)
    token_ids.append(XLM_SEP_ID)

    return token_ids


def camembert_tokens_pre_processor(token_ids: List[int]) -> List[int]:
    """Add camembert style special tokens.

    Args:
        token_ids: List of token ids without any special tokens.

    Returns:
        List of token ids augmented with special tokens.
    """
    CAMEMBERT_BEG_ID = 5
    CAMEMBERT_END_ID = 6

    token_ids.insert(0, CAMEMBERT_BEG_ID)
    token_ids.append(CAMEMBERT_END_ID)

    return token_ids


def bert_embeddings_post_processor(
    sequence_embeddings: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Post-process embeddings from BERT.

    by removing CLS and SEP embeddings and returning CLS token embedding as
    sentence representation.

    Args:
        sequence_embeddings: Sequence of token level embeddings received as output from
            BERT.

    Returns:
        sentence level embedding and post-processed sequence level embedding.
    """
    sentence_embedding = sequence_embeddings[0]
    post_processed_embedding = sequence_embeddings[1:-1]

    return sentence_embedding, post_processed_embedding


def gpt_embeddings_post_processor(
    sequence_embeddings: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Post-process embeddings from GPT models.

    by taking a mean over sequence embeddings and returning that as sentence
    representation.

    Args:
        sequence_embeddings: Sequence of token level embeddings received as output from
            GPT.

    Returns:
        sentence level embedding and post-processed sequence level embedding.
    """
    sentence_embedding = np.mean(sequence_embeddings, axis=0)
    post_processed_embedding = sequence_embeddings

    return sentence_embedding, post_processed_embedding


def xlnet_embeddings_post_processor(
    sequence_embeddings: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Post-process embeddings from XLNet models.

    by taking a mean over sequence embeddings and returning that as sentence
    representation. Remove last two time steps corresponding
    to special tokens from the sequence embeddings.

    Args:
        sequence_embeddings: Sequence of token level embeddings received as output from
            XLNet.

    Returns:
        sentence level embedding and post-processed sequence level embedding.
    """
    post_processed_embedding = sequence_embeddings[:-2]
    sentence_embedding = np.mean(post_processed_embedding, axis=0)

    return sentence_embedding, post_processed_embedding


def roberta_embeddings_post_processor(
    sequence_embeddings: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Post process embeddings from Roberta models.

    by taking a mean over sequence embeddings and returning that as sentence
    representation. Remove first and last time steps
    corresponding to special tokens from the sequence embeddings.

    Args:
        sequence_embeddings: Sequence of token level embeddings received as output from
            Roberta

    Returns:
        sentence level embedding and post-processed sequence level embedding
    """
    post_processed_embedding = sequence_embeddings[1:-1]
    sentence_embedding = np.mean(post_processed_embedding, axis=0)

    return sentence_embedding, post_processed_embedding


def xlm_embeddings_post_processor(
    sequence_embeddings: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """Post process embeddings from XLM models.

    by taking a mean over sequence embeddings and returning that as sentence
    representation. Remove first and last time steps
    corresponding to special tokens from the sequence embeddings.

    Args:
        sequence_embeddings: Sequence of token level embeddings received as output from
            XLM

    Returns:
        sentence level embedding and post-processed sequence level embedding
    """
    post_processed_embedding = sequence_embeddings[1:-1]
    sentence_embedding = np.mean(post_processed_embedding, axis=0)

    return sentence_embedding, post_processed_embedding


def bert_tokens_cleaner(
    token_ids: List[int], token_strings: List[Text]
) -> Tuple[List[int], List[Text]]:
    """Token cleanup method for BERT.

    Clean up tokens with the extra delimiters(##) BERT adds while breaking a token into
    sub-tokens.

    Args:
        token_ids: List of token ids received as output from BERT Tokenizer.
        token_strings: List of token strings received as output from BERT Tokenizer.

    Returns:
        Cleaned token ids and token strings.
    """
    return cleanup_tokens(list(zip(token_ids, token_strings)), "##")


def openaigpt_tokens_cleaner(
    token_ids: List[int], token_strings: List[Text]
) -> Tuple[List[int], List[Text]]:
    """Token cleanup method for GPT.

    Clean up tokens with the extra delimiters(</w>) OpenAIGPT adds while breaking a
    token into sub-tokens.

    Args:
        token_ids: List of token ids received as output from GPT Tokenizer.
        token_strings: List of token strings received as output from GPT Tokenizer.

    Returns:
        Cleaned token ids and token strings.
    """
    return cleanup_tokens(list(zip(token_ids, token_strings)), "</w>")


def gpt2_tokens_cleaner(
    token_ids: List[int], token_strings: List[Text]
) -> Tuple[List[int], List[Text]]:
    """Token cleanup method for GPT2.

    Clean up tokens with the extra delimiters(Ġ) GPT2 adds while breaking a token into
    sub-tokens.

    Args:
        token_ids: List of token ids received as output from GPT Tokenizer.
        token_strings: List of token strings received as output from GPT Tokenizer.

    Returns:
        Cleaned token ids and token strings.
    """
    return cleanup_tokens(list(zip(token_ids, token_strings)), "Ġ")


def xlnet_tokens_cleaner(
    token_ids: List[int], token_strings: List[Text]
) -> Tuple[List[int], List[Text]]:
    """Token cleanup method for XLNet.

    Clean up tokens with the extra delimiters(▁) XLNet adds while breaking a token into
    sub-tokens.

    Args:
        token_ids: List of token ids received as output from GPT Tokenizer.
        token_strings: List of token strings received as output from GPT Tokenizer.

    Returns:
        Cleaned token ids and token strings.
    """
    return cleanup_tokens(list(zip(token_ids, token_strings)), "▁")



from typing import TYPE_CHECKING, List, Optional, Text, Union

if TYPE_CHECKING:
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.generator import TrackerWithCachedStates
    from rasa.shared.core.training_data.structures import StoryGraph
    from rasa.shared.importers.importer import TrainingDataImporter


def extract_story_graph(
    resource_name: Text, domain: "Domain", exclusion_percentage: Optional[int] = None
) -> "StoryGraph":
    """Loads training stories / rules from file or directory.

    Args:
        resource_name: Path to file or directory.
        domain: The model domain.
        exclusion_percentage: Percentage of stories which should be dropped. `None`
            if all training data should be used.

    Returns:
        The loaded training data as graph.
    """
    import rasa.shared.core.training_data.loading as core_loading
    from rasa.shared.core.training_data.structures import StoryGraph

    story_steps = core_loading.load_data_from_resource(
        resource_name, domain, exclusion_percentage=exclusion_percentage
    )
    return StoryGraph(story_steps)


def load_data(
    resource_name: Union[Text, "TrainingDataImporter"],
    domain: "Domain",
    remove_duplicates: bool = True,
    unique_last_num_states: Optional[int] = None,
    augmentation_factor: int = 50,
    tracker_limit: Optional[int] = None,
    use_story_concatenation: bool = True,
    debug_plots: bool = False,
    exclusion_percentage: Optional[int] = None,
) -> List["TrackerWithCachedStates"]:
    """Load training data from a resource.

    Args:
        resource_name: resource to load the data from. either a path or an importer
        domain: domain used for loading
        remove_duplicates: should duplicated training examples be removed?
        unique_last_num_states: number of states in a conversation that make the
            a tracker unique (this is used to identify duplicates)
        augmentation_factor:
            by how much should the story training data be augmented
        tracker_limit:
            maximum number of trackers to generate during augmentation
        use_story_concatenation:
            should stories be concatenated when doing data augmentation
        debug_plots:
            generate debug plots during loading
        exclusion_percentage:
            how much data to exclude

    Returns:
        list of loaded trackers
    """
    from rasa.shared.core.generator import TrainingDataGenerator
    from rasa.shared.importers.importer import TrainingDataImporter

    if resource_name:
        if isinstance(resource_name, TrainingDataImporter):
            graph = resource_name.get_stories(exclusion_percentage=exclusion_percentage)
        else:
            graph = extract_story_graph(
                resource_name, domain, exclusion_percentage=exclusion_percentage
            )

        g = TrainingDataGenerator(
            graph,
            domain,
            remove_duplicates,
            unique_last_num_states,
            augmentation_factor,
            tracker_limit,
            use_story_concatenation,
            debug_plots,
        )
        return g.generate()
    else:
        return []



import asyncio
import importlib
import json
from contextlib import asynccontextmanager
from typing import Any, Dict, List, Optional, Union

import openai
import structlog
from jinja2 import Template
from typing_extensions import AsyncGenerator

from rasa.builder import config
from rasa.builder.config import COPILOT_DOCUMENTATION_SEARCH_QUERY_HISTORY_MESSAGES
from rasa.builder.copilot.constants import (
    COPILOT_LAST_USER_MESSAGE_CONTEXT_PROMPT_FILE,
    COPILOT_PROMPTS_DIR,
    COPILOT_PROMPTS_FILE,
    COPILOT_TRAINING_ERROR_HANDLER_PROMPT_FILE,
    ROLE_COPILOT,
    ROLE_COPILOT_INTERNAL,
    ROLE_USER,
)
from rasa.builder.copilot.exceptions import CopilotStreamError
from rasa.builder.copilot.models import (
    ChatMessage,
    CopilotChatMessage,
    CopilotContext,
    CopilotGenerationContext,
    CopilotSystemMessage,
    EventContent,
    FileContent,
    InternalCopilotRequestChatMessage,
    ResponseCategory,
    UsageStatistics,
    UserChatMessage,
)
from rasa.builder.document_retrieval.inkeep_document_retrieval import (
    InKeepDocumentRetrieval,
)
from rasa.builder.document_retrieval.models import Document
from rasa.builder.exceptions import (
    DocumentRetrievalError,
)
from rasa.builder.shared.tracker_context import TrackerContext
from rasa.shared.constants import PACKAGE_NAME

structlogger = structlog.get_logger()


class Copilot:
    def __init__(self) -> None:
        self._inkeep_document_retrieval = InKeepDocumentRetrieval()

        # Load both prompt templates up-front
        self._system_message_prompt_template = Template(
            importlib.resources.read_text(
                f"{PACKAGE_NAME}.{COPILOT_PROMPTS_DIR}",
                COPILOT_PROMPTS_FILE,
            )
        )
        self._last_user_message_context_prompt_template = Template(
            importlib.resources.read_text(
                f"{PACKAGE_NAME}.{COPILOT_PROMPTS_DIR}",
                COPILOT_LAST_USER_MESSAGE_CONTEXT_PROMPT_FILE,
            )
        )
        self._training_error_handler_prompt_template = Template(
            importlib.resources.read_text(
                f"{PACKAGE_NAME}.{COPILOT_PROMPTS_DIR}",
                COPILOT_TRAINING_ERROR_HANDLER_PROMPT_FILE,
            )
        )

        # The final stream chunk includes usage statistics.
        self.usage_statistics = UsageStatistics()

    @asynccontextmanager
    async def _get_client(self) -> AsyncGenerator[openai.AsyncOpenAI, None]:
        """Create a fresh OpenAI client, yield it, and always close it."""
        client = openai.AsyncOpenAI(timeout=config.OPENAI_TIMEOUT)
        try:
            yield client
        except Exception as e:
            structlogger.error("copilot.llm_client_error", error=str(e))
            raise
        finally:
            try:
                await client.close()
            except Exception as exc:
                # Closing should not break request processing, but we log it
                structlogger.warning(
                    "copilot.llm_client_close_error",
                    event_info="Failed to close OpenAI client cleanly.",
                    error=str(exc),
                )

    async def search_rasa_documentation(
        self,
        context: CopilotContext,
    ) -> List[Document]:
        """Search Rasa documentation for relevant information.

        Args:
            context: The context of the copilot.

        Returns:
            A list of Document objects. Empty list is returned if the search fails.
        """
        try:
            query = self._create_documentation_search_query(context)
            return await self._inkeep_document_retrieval.retrieve_documents(query)
        except DocumentRetrievalError as e:
            structlogger.error(
                "copilot.search_rasa_documentation.error",
                event_info=(
                    f"Copilot: Searching Rasa documentation for query '{query}' "
                    f"failed with the following error: {e}. Returning empty list."
                ),
                query=query,
                error=str(e),
            )
            return []

    async def generate_response(
        self,
        context: CopilotContext,
    ) -> tuple[AsyncGenerator[str, None], CopilotGenerationContext]:
        """Generate a response from the copilot.

        This method performs document retrieval and response generation as a single
        atomic operation. The returned documents are the supporting evidence used
        to generate the response, ensuring consistency between the response content
        and its sources.

        Args:
            context: The context of the copilot.

        Returns:
            A tuple containing the async response stream and a
            CopilotGenerationContext object with relevant documents, and all the
            messages used to generate the response.

        Raises:
            CopilotStreamError: If the stream fails.
            Exception: If an unexpected error occurs.
        """
        relevant_documents = await self.search_rasa_documentation(context)
        messages = await self._build_messages(context, relevant_documents)
        tracker_event_attachments = self._extract_tracker_event_attachments(
            context.copilot_chat_history[-1]
        )

        support_evidence = CopilotGenerationContext(
            relevant_documents=relevant_documents,
            system_message=messages[0],
            chat_history=messages[1:-1],
            last_user_message=messages[-1],
            tracker_event_attachments=tracker_event_attachments,
        )

        return (
            self._stream_response(messages),
            support_evidence,
        )

    async def _stream_response(
        self, messages: List[Dict[str, Any]]
    ) -> AsyncGenerator[str, None]:
        """Stream markdown chunks one by one."""
        self.usage_statistics.reset()

        try:
            async with self._get_client() as client:
                stream = await client.chat.completions.create(
                    model=config.OPENAI_MODEL,
                    messages=messages,  # type: ignore
                    temperature=config.OPENAI_TEMPERATURE,
                    stream=True,
                    stream_options={"include_usage": True},
                )
                async for chunk in stream:
                    # The final chunk, which contains the usage statistics,
                    # arrives with an empty `choices` list.
                    if not chunk.choices:
                        self.usage_statistics.update_from_stream_chunk(chunk)
                        # Nothing to yield – continue to the next chunk.
                        continue

                    delta = chunk.choices[0].delta
                    if delta and delta.content:
                        yield delta.content
        except openai.OpenAIError as e:
            structlogger.exception("copilot.stream_response.api_error", error=str(e))
            raise CopilotStreamError(
                "Failed to stream response from OpenAI API."
            ) from e
        except asyncio.TimeoutError as e:
            structlogger.exception(
                "copilot.stream_response.timeout_error", error=str(e)
            )
            raise CopilotStreamError("Request to OpenAI API timed out.") from e
        except Exception as e:
            structlogger.exception(
                "copilot.stream_response.unexpected_error", error=str(e)
            )
            raise

    async def _build_messages(
        self,
        context: CopilotContext,
        relevant_documents: List[Document],
    ) -> List[Dict[str, Any]]:
        """Build the complete message list for the OpenAI API.

        Args:
            context: The context of the copilot.
            relevant_documents: The relevant documents to use in the context.

        Returns:
            A list of messages in OpenAI format.
        """
        if not context.copilot_chat_history:
            return []

        past_messages = self._create_chat_history_messages(
            context.copilot_chat_history[:-1]
        )

        latest_message = self._process_latest_message(
            context.copilot_chat_history[-1], context, relevant_documents
        )
        system_message = self._create_system_message()

        return [system_message, *past_messages, latest_message]

    def _create_chat_history_messages(
        self, chat_history: List[Union[UserChatMessage, CopilotChatMessage]]
    ) -> List[Dict[str, Any]]:
        """Filter and convert past messages to OpenAI format.

        Excludes guardrails policy violations and non-user/copilot messages.

        Args:
            chat_history: List of chat messages to filter and convert.

        Returns:
            List of messages in OpenAI format
        """
        filtered_messages = []

        for message in chat_history:
            if (
                message.response_category
                != ResponseCategory.GUARDRAILS_POLICY_VIOLATION
                and message.role in [ROLE_USER, ROLE_COPILOT]
            ):
                filtered_messages.append(message)

        return [message.build_openai_message() for message in filtered_messages]

    def _process_latest_message(
        self,
        latest_message: Any,
        context: CopilotContext,
        relevant_documents: List[Document],
    ) -> Dict[str, Any]:
        """Process the latest message and convert it to OpenAI format.

        Args:
            latest_message: The most recent message from the chat history.
            context: The copilot context containing conversation state.
            relevant_documents: List of relevant documents for context.

        Returns:
            Message in OpenAI format.

        Raises:
            ValueError: If the message type is not supported.
        """
        if isinstance(latest_message, UserChatMessage):
            tracker_event_attachments = latest_message.get_content_blocks_by_type(
                EventContent
            )
            rendered_prompt = self._render_last_user_message_context_prompt(
                context, relevant_documents, tracker_event_attachments
            )
            return latest_message.build_openai_message(prompt=rendered_prompt)

        elif isinstance(latest_message, InternalCopilotRequestChatMessage):
            rendered_prompt = self._render_training_error_handler_prompt(
                latest_message, relevant_documents
            )
            return latest_message.build_openai_message(prompt=rendered_prompt)

        else:
            raise ValueError(f"Unexpected message type: {type(latest_message)}")

    def _create_system_message(self) -> Dict[str, Any]:
        """Create the system message for the conversation.

        Returns:
            System message in OpenAI format with rendered prompt template.
        """
        rendered_prompt = self._system_message_prompt_template.render()
        return CopilotSystemMessage().build_openai_message(prompt=rendered_prompt)

    def _render_last_user_message_context_prompt(
        self,
        context: CopilotContext,
        relevant_documents: List[Document],
        tracker_event_attachments: List[EventContent],
    ) -> str:
        # Format relevant documentation
        documents = [doc.model_dump() for doc in relevant_documents]
        # Format conversation history
        conversation = self._format_conversation_history(context.tracker_context)
        # Format current state
        current_state = self._format_current_state(context.tracker_context)
        # Format tracker events
        attachments = self._format_tracker_event_attachments(tracker_event_attachments)

        rendered_prompt = self._last_user_message_context_prompt_template.render(
            current_conversation=conversation,
            current_state=current_state,
            assistant_logs=context.assistant_logs,
            assistant_files=context.assistant_files,
            documentation_results=documents,
            attachments=attachments,
        )
        return rendered_prompt

    def _render_training_error_handler_prompt(
        self,
        internal_request_message: InternalCopilotRequestChatMessage,
        relevant_documents: List[Document],
    ) -> str:
        """Render the training error handler prompt with documentation and context.

        Args:
            internal_request_message: Internal request message.
            context: The copilot context.
            relevant_documents: List of relevant documents for context.

        Returns:
            Rendered prompt string for training error analysis.
        """
        modified_files_dicts: Dict[str, str] = {
            file.file_path: file.file_content
            for file in internal_request_message.get_content_blocks_by_type(FileContent)
        }
        rendered_prompt = self._training_error_handler_prompt_template.render(
            logs=internal_request_message.get_flattened_log_content(),
            modified_files=modified_files_dicts,
            documentation_results=self._format_documents(relevant_documents),
        )

        return rendered_prompt

    @staticmethod
    def _create_documentation_search_query(context: CopilotContext) -> str:
        """Format chat messages between user and copilot for documentation search.

        Filters out guardrails policy violations and only includes messages with
        USER or COPILOT roles, then takes the last N relevant messages.
        """
        role_to_prefix = {
            ROLE_USER: "User",
            ROLE_COPILOT: "Assistant",
            ROLE_COPILOT_INTERNAL: "User",
        }
        allowed_message_types = (
            UserChatMessage,
            InternalCopilotRequestChatMessage,
            CopilotChatMessage,
        )

        query_chat_history: List[str] = []

        for message in reversed(context.copilot_chat_history):
            if (
                message.response_category
                == ResponseCategory.GUARDRAILS_POLICY_VIOLATION
                or not isinstance(message, allowed_message_types)
            ):
                continue

            if (
                len(query_chat_history)
                >= COPILOT_DOCUMENTATION_SEARCH_QUERY_HISTORY_MESSAGES
            ):
                break

            prefix = role_to_prefix[message.role]
            text = (
                Copilot._format_internal_message_for_query_chat_history(message)
                if isinstance(message, InternalCopilotRequestChatMessage)
                else Copilot._format_normal_message_for_query_chat_history(message)
            )
            query_chat_history.insert(0, f"{prefix}: {text}")

        return "\n".join(query_chat_history)

    @staticmethod
    def _format_documents(results: List[Document]) -> Optional[str]:
        """Format documentation search results as JSON dump to be used in the prompt."""
        # We want the special message that indicates no relevant documentation source
        # found if there are no results.
        if not results:
            return None

        formatted_results: Dict[str, Any] = {
            "sources": [
                {
                    # Start the reference from 1, not 0.
                    "idx": idx + 1,
                    "title": result.title,
                    "url": result.url,
                    "content": result.content,
                }
                for idx, result in enumerate(results)
            ]
        }
        return json.dumps(formatted_results, ensure_ascii=False, indent=2)

    @staticmethod
    def _format_conversation_history(tracker_context: Optional[TrackerContext]) -> str:
        """Format conversation history from TrackerContext using nested turn structure.

        Args:
            tracker_context: The TrackerContext containing conversation data.

        Returns:
            A JSON string with turns containing user_input, assistant_response,
            and context.

        Example:
            ```json
            {
                "conversation_history": [
                    {
                        "turn_id": 1,
                        "USER": {
                            "text": "I want to transfer money",
                            "predicted_commands": ["start flow", "set slot", ...]
                        },
                        "BOT": [
                            {"text": "How much would you like to transfer?"}
                        ],
                        "other_tracker_events": [
                            {
                                "event": "action_executed",
                                "data": {"action_name": "action_ask_amount"}
                            },
                            {
                                "event": "slot_set",
                                "data": {
                                    "slot_name": "amount_of_money",
                                    "slot_value": 100,
                                },
                            }
                        ]
                    }
                ]
            }
            ```
        """
        conversation_history: Dict[str, Any] = {
            "conversation_history": [],
        }

        if not tracker_context or not tracker_context.conversation_turns:
            return json.dumps(conversation_history, ensure_ascii=False, indent=2)

        conversation_turns: List[Dict[str, Any]] = []
        user_prefix = "USER"
        assistant_prefix = "BOT"

        for turn_idx, turn in enumerate(tracker_context.conversation_turns, 1):
            turn_data: Dict[str, Any] = {"turn_id": turn_idx}

            # Add user if present
            if turn.user_message:
                turn_data[user_prefix] = {
                    "text": turn.user_message.text,
                    "predicted_commands": turn.user_message.predicted_commands,
                }

            # Add assistant messages if present
            if turn.assistant_messages:
                turn_data[assistant_prefix] = [
                    {"text": assistant_message.text}
                    for assistant_message in turn.assistant_messages
                ]

            # Add other tracker events
            if turn.context_events:
                other_events = [event.model_dump() for event in turn.context_events]
                turn_data["other_tracker_events"] = other_events

            conversation_turns.append(turn_data)

        conversation_history["conversation_history"] = conversation_turns
        return json.dumps(conversation_history, ensure_ascii=False, indent=2)

    @staticmethod
    def _format_current_state(tracker_context: Optional[TrackerContext]) -> str:
        """Format current state from TrackerContext for LLM consumption.

        Args:
            tracker_context: The TrackerContext containing current state data.

        Returns:
            A JSON string containing the current state information.
        """
        if not tracker_context or not tracker_context.current_state:
            return json.dumps({}, ensure_ascii=False, indent=2)
        current_state = tracker_context.current_state.model_dump()
        return json.dumps(current_state, ensure_ascii=False, indent=2)

    @staticmethod
    def _format_normal_message_for_query_chat_history(
        message: Union[UserChatMessage, CopilotChatMessage],
    ) -> str:
        """Format normal message for query chat history."""
        return f"{message.get_flattened_text_content()}"

    @staticmethod
    def _format_internal_message_for_query_chat_history(
        message: InternalCopilotRequestChatMessage,
    ) -> str:
        """Format internal copilot request message for query chat history."""
        text_content = message.get_flattened_text_content()
        log_content = message.get_flattened_log_content()
        if text_content and log_content:
            return f"{text_content}\nLogs: {log_content}"
        elif text_content:
            return text_content
        elif log_content:
            return f"Logs: {log_content}"
        else:
            return ""

    @staticmethod
    def _format_tracker_event_attachments(events: List[EventContent]) -> Optional[str]:
        """Format tracker events as JSON dump to be used in the prompt."""
        # We don't want to display the attachment sectin in the last user message
        # context prompt if there are no attachments.
        if not events:
            return None
        # If there are attachments, return the formatted JSON dump.
        return json.dumps(
            [event_content.model_dump() for event_content in events],
            ensure_ascii=False,
            indent=2,
        )

    @staticmethod
    def _extract_tracker_event_attachments(message: ChatMessage) -> List[EventContent]:
        """Extract the tracker event attachments from the message."""
        if not isinstance(message, UserChatMessage):
            return []
        return message.get_content_blocks_by_type(EventContent)



# A central place to define all environment variables used by Rasa

AUTH_TOKEN_ENV = "AUTH_TOKEN"
JWT_SECRET_ENV = "JWT_SECRET"
JWT_METHOD_ENV = "JWT_METHOD"
DEFAULT_JWT_METHOD = "HS256"
JWT_PRIVATE_KEY_ENV = "JWT_PRIVATE_KEY"

REMOTE_STORAGE_PATH_ENV = "REMOTE_STORAGE_PATH"
BUCKET_NAME_ENV = "BUCKET_NAME"
AWS_ENDPOINT_URL_ENV = "AWS_ENDPOINT_URL"
AZURE_CONTAINER_ENV = "AZURE_CONTAINER"
AZURE_ACCOUNT_NAME_ENV = "AZURE_ACCOUNT_NAME"
AZURE_ACCOUNT_KEY_ENV = "AZURE_ACCOUNT_KEY"



import asyncio
import base64
import copy
import json
import logging
import warnings
from dataclasses import dataclass
from functools import wraps
from typing import Any, Awaitable, Callable, Dict, Optional, Text

import structlog

import rasa.shared.utils.io
from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage
from rasa.utils.beta import ensure_beta_feature_is_enabled

# ignore ResourceWarning, InsecureRequestWarning
warnings.filterwarnings("ignore", category=ResourceWarning)

from cvg_sdk.api.call_api import CallApi  # noqa: E402
from cvg_sdk.api_client import ApiClient  # noqa: E402
from cvg_sdk.configuration import Configuration  # noqa: E402
from cvg_sdk.model.outbound_call_result import OutboundCallResult  # noqa: E402
from cvg_sdk.model.say_parameters import SayParameters  # noqa: E402
from sanic import Blueprint, response  # noqa: E402
from sanic.request import Request  # noqa: E402
from sanic.response import HTTPResponse  # noqa: E402

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()

CHANNEL_NAME = "vier_cvg"
FEATURE_NAME = "VIER Voice Connector"
RASA_PRO_BETA_VIER = "RASA_PRO_BETA_VIER"


def make_metadata(payload: Any) -> Dict[Text, Any]:
    return {"cvg_body": payload}


@dataclass
class Recipient:
    dialog_id: str
    project_token: str
    reseller_token: str


def parse_recipient_id(recipient_id: Text) -> Recipient:
    parsed_json = json.loads(
        base64.b64decode(bytes(recipient_id, "utf-8")).decode("utf-8")
    )
    project_context = parsed_json["projectContext"]
    return Recipient(
        parsed_json["dialogId"],
        project_context["projectToken"],
        project_context["resellerToken"],
    )


class CVGOutput(OutputChannel):
    """Output channel for the Cognitive Voice Gateway."""

    on_message: Callable[[UserMessage], Awaitable[Any]]

    @classmethod
    def name(cls) -> Text:
        return CHANNEL_NAME

    def __init__(
        self,
        callback_base_url: Text,
        on_message: Callable[[UserMessage], Awaitable[Any]],
        proxy: Optional[Text] = None,
    ) -> None:
        configuration = Configuration.get_default_copy()
        configuration.host = callback_base_url
        configuration.proxy = proxy
        configuration.verify_ssl = False

        self.on_message = on_message

        self.api_client = ApiClient(configuration=configuration)
        self.call_api = CallApi(self.api_client)

    async def send_text_message(
        self,
        recipient_id: Text,
        text: Text,
        **kwargs: Any,
    ) -> None:
        structlogger.debug("vier_cvg.send.text.message", text=copy.deepcopy(text))

        dialog_id = parse_recipient_id(recipient_id).dialog_id
        self.call_api.say(SayParameters(dialog_id=dialog_id, text=text))

    async def _execute_operation_by_name(
        self,
        operation_name: Text,
        body: Any,
        recipient_id: Text,
    ) -> Any:
        recipient = parse_recipient_id(recipient_id)
        dialog_id = recipient.dialog_id
        reseller_token = recipient.reseller_token

        async def handle_outbound_call_result(
            outbound_call_result: OutboundCallResult,
        ) -> Any:
            if outbound_call_result.status == "Success":
                user_message = UserMessage(
                    text="/cvg_outbound_success",
                    output_channel=self,
                    sender_id=recipient_id,
                    input_channel=CHANNEL_NAME,
                    metadata=make_metadata(outbound_call_result.to_dict()),
                )
            elif outbound_call_result.status == "Failure":
                user_message = UserMessage(
                    text="/cvg_outbound_failure",
                    output_channel=self,
                    sender_id=recipient_id,
                    input_channel=CHANNEL_NAME,
                    metadata=make_metadata(outbound_call_result.to_dict()),
                )
            else:
                return response.text(
                    f"Invalid OutboundCallResult status: {outbound_call_result.status}",
                    status=400,
                )

            logger.info(
                "Creating incoming UserMessage: {output_channel=%s, sender_id=%s, metadata=%s}"  # noqa: E501
                % (
                    user_message.output_channel,
                    user_message.sender_id,
                    user_message.metadata,
                )
            )
            await self.on_message(user_message)

        newBody = copy.deepcopy(body)
        if newBody is None:
            newBody = {}
        path = operation_name[3:].replace("_", "/")
        if operation_name.startswith("cvg_call_"):
            if "dialogId" not in body:
                newBody["dialogId"] = dialog_id

            # The response from forward and bridge must be handled
            handle_result_outbound_call_result_for = [
                "cvg_call_forward",
                "cvg_call_bridge",
            ]
            if operation_name in handle_result_outbound_call_result_for:
                # The request must be async: We cannot trigger another intent,
                # while the send_ function of OutputChannel is not finished yet.
                # (conversation is locked)
                self.api_client.pool.apply_async(
                    self.api_client.call_api,
                    (path, "POST"),
                    {"body": newBody, "response_type": (OutboundCallResult,)},
                    callback=lambda result: asyncio.run(
                        handle_outbound_call_result(result[0])
                    ),
                )

            return self.api_client.call_api(path, "POST", body=newBody)
        elif operation_name.startswith("cvg_dialog_"):
            if operation_name == "cvg_dialog_delete":
                return self.api_client.call_api(
                    f"/dialog/{reseller_token}/{dialog_id}", "DELETE", body=newBody
                )
            elif operation_name == "cvg_dialog_data":
                return self.api_client.call_api(
                    f"/dialog/{reseller_token}/{dialog_id}/data", "POST", body=newBody
                )
            else:
                logger.error(
                    f"Dialog operation {operation_name} not found/not implemented yet. "
                    f"Please consider using the cvg-python-sdk in one of your actions."
                )
        else:
            logger.error(f"Operation {operation_name} not found/not implemented yet")
        logger.info("Ran operation: " + operation_name)

    async def send_custom_json(
        self,
        recipient_id: Text,
        json_message: Dict[Text, Any],
        **kwargs: Any,
    ) -> None:
        structlogger.debug(
            "vier_cvg.send.json",
            json_message=copy.deepcopy(json.dumps(json_message)),
            recipient_id=recipient_id,
        )
        for operation_name, body in json_message.items():
            await self._execute_operation_by_name(operation_name, body, recipient_id)

    async def send_image_url(*args: Any, **kwargs: Any) -> None:
        # We do not support images.
        rasa.shared.utils.io.raise_warning(
            "Ignoring image URL."
            "We cannot represent images as a voice bot."
            "Please define a voice-friendly alternative."
        )


class CVGInput(InputChannel):
    """Input channel for the Cognitive Voice Gateway."""

    @classmethod
    def name(cls) -> Text:
        return CHANNEL_NAME

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if credentials is None:
            cls.raise_missing_credentials_exception()
        token = credentials.get("token")
        if token is None or isinstance(token, str) or len(token) == 0:
            raise ValueError(
                "No authentication token has been configured in your credentials.yml!"
            )
        proxy = credentials.get("proxy")
        start_intent = credentials.get("start_intent")
        if start_intent is None:
            start_intent = "/cvg_session"

        return cls(token, start_intent, proxy)

    def __init__(
        self, token: Text, start_intent: Text, proxy: Optional[Text] = None
    ) -> None:
        ensure_beta_feature_is_enabled(
            feature_name=FEATURE_NAME,
            env_flag=RASA_PRO_BETA_VIER,
        )

        self.callback = None
        self.expected_authorization_header_value = f"Bearer {token}"
        self.proxy = proxy
        self.start_intent = start_intent

    async def process_message(
        self,
        request: Request,
        on_new_message: Callable[[UserMessage], Awaitable[Any]],
        text: Text,
        sender_id: Optional[Text],
    ) -> Any:
        try:
            if text[-1] == ".":
                text = text[:-1]

            metadata = make_metadata(request.json)
            user_msg = UserMessage(
                text=text,
                output_channel=CVGOutput(
                    request.json["callback"], on_new_message, self.proxy
                ),
                sender_id=sender_id,
                input_channel=CHANNEL_NAME,
                metadata=metadata,
            )

            structlogger.debug(
                "vier_cvg.process.message",
                text=copy.deepcopy(text),
                output_channel=user_msg.output_channel,
                sender_id=sender_id,
                metadata=metadata,
            )

            await on_new_message(user_msg)
        except Exception as e:
            logger.error(f"Exception when trying to handle message: {e}")
            logger.error(str(e), exc_info=True)

        return response.text("")

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        def valid_request(func: Any) -> Any:
            def decorator(f: Any) -> Any:
                @wraps(f)
                async def decorated_function(
                    request: HTTPResponse, *args: Any, **kwargs: Any
                ) -> Any:
                    if (
                        request.headers.get("authorization")
                        != self.expected_authorization_header_value
                    ):
                        return response.text("bot token is invalid!", status=401)

                    if not request.headers.get("content-type") == "application/json":
                        return response.text(
                            "content-type is not supported. "
                            "Please use application/json",
                            status=415,
                        )
                    json_body = request.json  # type: ignore[attr-defined]
                    if json_body is None:
                        return response.text("body is not valid json.", status=400)
                    if json_body["dialogId"] is None:
                        return response.text("dialogId is required", status=400)
                    if json_body["callback"] is None:
                        return response.text("callback is required", status=400)
                    if json_body["projectContext"] is None:
                        return response.text("projectContext is required", status=400)
                    return await f(request, *args, **kwargs)

                return decorated_function

            return decorator(func)

        async def process_message_oneline(request: Request, text: Text) -> Any:
            sender_id_json = json.dumps(
                {
                    "dialogId": request.json["dialogId"],
                    "projectContext": request.json["projectContext"],
                }
            )
            sender_id_base64 = base64.b64encode(bytes(sender_id_json, "utf-8")).decode(
                "utf-8"
            )

            return await self.process_message(
                request,
                on_new_message,
                text=text,
                sender_id=sender_id_base64,
            )

        cvg_webhook = Blueprint(
            "vier_cvg_webhook",
            __name__,
        )

        @cvg_webhook.post("/session")
        @valid_request
        async def session(request: Request) -> HTTPResponse:
            return await process_message_oneline(request, self.start_intent)

        @cvg_webhook.post("/message")
        @valid_request
        async def message(request: Request) -> HTTPResponse:
            return await process_message_oneline(request, request.json["text"])

        @cvg_webhook.post("/answer")
        @valid_request
        async def answer(request: Request) -> HTTPResponse:
            return await process_message_oneline(
                request, "/cvg_answer_" + request.json["type"]["name"].lower()
            )

        @cvg_webhook.post("/inactivity")
        @valid_request
        async def inactivity(request: Request) -> HTTPResponse:
            return await process_message_oneline(request, "/cvg_inactivity")

        @cvg_webhook.post("/terminated")
        @valid_request
        async def terminated(request: Request) -> HTTPResponse:
            return await process_message_oneline(request, "/cvg_terminated")

        @cvg_webhook.post("/recording")
        @valid_request
        async def recording(request: Request) -> HTTPResponse:
            return await process_message_oneline(request, "/cvg_recording")

        return cvg_webhook



"""Service for generating Rasa projects from prompts."""

import json
import os
import shutil
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Generator, List, Optional

import structlog

from rasa.builder import config
from rasa.builder.exceptions import ProjectGenerationError, ValidationError
from rasa.builder.llm_service import get_skill_generation_messages, llm_service
from rasa.builder.logging_utils import capture_exception_with_context
from rasa.builder.models import BotFiles
from rasa.builder.project_info import ProjectInfo, ensure_first_used, load_project_info
from rasa.builder.template_cache import copy_cache_for_template_if_available
from rasa.builder.training_service import TrainingInput
from rasa.builder.validation_service import validate_project
from rasa.cli.scaffold import ProjectTemplateName, create_initial_project
from rasa.shared.constants import DEFAULT_MODELS_PATH
from rasa.shared.core.flows import yaml_flows_io
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.yaml import dump_obj_as_yaml_to_string
from rasa.utils.io import InvalidPathException, subpath

structlogger = structlog.get_logger()


class ProjectGenerator:
    """Service for generating Rasa projects from skill descriptions."""

    def __init__(self, project_folder: str) -> None:
        """Initialize the project generator with a folder for file persistence.

        Args:
            project_folder: Path to the folder where project files will be stored
        """
        self.project_folder = Path(project_folder)
        self.project_folder.mkdir(parents=True, exist_ok=True)

    @property
    def project_info(self) -> ProjectInfo:
        """Get the project info."""
        return load_project_info(self.project_folder)

    def is_empty(self) -> bool:
        """Check if the project folder is empty.

        Excluding hidden paths.
        """
        return not any(
            file.is_file()
            for file in self.project_folder.iterdir()
            if not file.name.startswith(".")
        )

    async def init_from_template(self, template: ProjectTemplateName) -> None:
        """Create the initial project files."""
        self.cleanup()
        create_initial_project(self.project_folder.as_posix(), template)
        # If a local cache for this template exists, copy it into the project.
        # We no longer download here to avoid blocking project creation.
        await copy_cache_for_template_if_available(template, self.project_folder)
        # needs to happen after caching, as we download/copy .rasa and that would
        # overwrite the project info file in .rasa
        ensure_first_used(self.project_folder)

    async def generate_project_with_retries(
        self,
        skill_description: str,
        template: ProjectTemplateName,
        max_retries: Optional[int] = None,
    ) -> Dict[str, Optional[str]]:
        """Generate a Rasa project with retry logic for validation failures.

        Args:
            skill_description: Natural language description of the skill
            rasa_config: Rasa configuration dictionary
            template: Project template to use for the initial project
            max_retries: Maximum number of retry attempts

        Returns:
            Dictionary of generated file contents (filename -> content)

        Raises:
            ProjectGenerationError: If generation fails after all retries
        """
        if max_retries is None:
            max_retries = config.MAX_RETRIES

        # if ever we do not use the template, we need to ensure first_used is set
        # separately!
        await self.init_from_template(template)

        project_data = self._get_bot_data_for_llm()

        initial_messages = get_skill_generation_messages(
            skill_description, project_data
        )

        async def _generate_with_retry(
            messages: List[Dict[str, Any]], attempts_left: int
        ) -> Dict[str, Optional[str]]:
            try:
                # Generate project data using LLM
                project_data = await llm_service.generate_rasa_project(messages)

                # Update stored bot data
                self._update_bot_files_from_llm_response(project_data)

                bot_files = self.get_bot_files()
                structlogger.info(
                    "project_generator.generated_project",
                    attempts_left=attempts_left,
                    files=list(bot_files.keys()),
                )

                # Validate the generated project
                await self._validate_generated_project()

                structlogger.info(
                    "project_generator.validation_success", attempts_left=attempts_left
                )

                return bot_files

            except ValidationError as e:
                structlogger.error(
                    "project_generator.validation_error",
                    error=str(e),
                    attempts_left=attempts_left,
                )

                if attempts_left <= 0:
                    raise ProjectGenerationError(
                        f"Failed to generate valid Rasa project: {e}", max_retries
                    )

                # Create error feedback for next attempt
                error_feedback_messages = messages + [
                    {
                        "role": "assistant",
                        "content": json.dumps(project_data),
                    },
                    {
                        "role": "user",
                        "content": dedent(f"""
                            Previous attempt failed validation with error: {e}

                            Please fix the issues and generate a valid Rasa project.
                            Pay special attention to:
                            - Proper YAML syntax
                            - Required fields in domain and flows
                            - Consistent naming between flows and domain
                            - Valid slot types and mappings
                        """).strip(),
                    },
                ]

                return await _generate_with_retry(
                    error_feedback_messages, attempts_left - 1
                )

            except Exception as e:
                structlogger.error(
                    "project_generator.generation_error",
                    error=str(e),
                    attempts_left=attempts_left,
                )

                if attempts_left <= 0:
                    raise ProjectGenerationError(
                        f"Failed to generate Rasa project: {e}", max_retries
                    )

                # For non-validation errors, retry with original messages
                return await _generate_with_retry(initial_messages, attempts_left - 1)

        return await _generate_with_retry(initial_messages, max_retries)

    async def _validate_generated_project(self) -> None:
        """Validate the generated project using the validation service."""
        importer = self._create_importer()
        validation_error = await validate_project(importer)

        if validation_error:
            raise ValidationError(validation_error)

    def _get_endpoints_file(self) -> Path:
        """Get the endpoints file."""
        return self.project_folder / "endpoints.yml"

    def _get_config_file(self) -> Path:
        """Get the config file."""
        return self.project_folder / "config.yml"

    def get_training_input(self) -> TrainingInput:
        """Get the training input."""
        return TrainingInput(
            importer=self._create_importer(),
            endpoints_file=self._get_endpoints_file(),
            config_file=self._get_config_file(),
        )

    def _create_importer(self) -> TrainingDataImporter:
        """Create a training data importer from the current bot files."""
        try:
            if (self.project_folder / "domain.yml").exists():
                domain_path = self.project_folder / "domain.yml"
            else:
                domain_path = self.project_folder / "domain"

            return TrainingDataImporter.load_from_config(
                config_path=str(self._get_config_file()),
                domain_path=str(domain_path),
                training_data_paths=[str(self.project_folder / "data")],
                args={},
            )

        except Exception as e:
            raise ValidationError(f"Failed to create importer: {e}")

    def get_bot_files(
        self,
        allowed_file_extensions: Optional[List[str]] = None,
        exclude_docs_directory: bool = False,
    ) -> BotFiles:
        """Get the current bot files by reading from disk.

        Args:
            allowed_file_extensions: Optional list of file extensions to include.
                If None, fetch all files. If provided, only fetch files with matching
                extensions. Use `""` empty string to allow files with no extensions.
            exclude_docs_directory: Optional boolean indicating whether to exclude.

        Returns:
            Dictionary of file contents with relative paths as keys
        """
        bot_files: BotFiles = {}

        for file in self.bot_file_paths():
            relative_path = file.relative_to(self.project_folder)

            # Exclude the docs directory if specified
            if exclude_docs_directory and relative_path.parts[0] == "docs":
                continue

            # Exclude the files by file extensions if specified
            if allowed_file_extensions is not None:
                allowed_file_extensions = [
                    ext.lower() for ext in allowed_file_extensions
                ]
                if file.suffix.lstrip(".").lower() not in allowed_file_extensions:
                    continue
            # Read file content and store with relative path as key
            try:
                bot_files[relative_path.as_posix()] = file.read_text(encoding="utf-8")
            except Exception as e:
                structlogger.debug(
                    "project_generator.get_bot_files.error",
                    error=str(e),
                    file_path=file.as_posix(),
                )
                bot_files[relative_path.as_posix()] = None
        return bot_files

    def is_restricted_path(self, path: Path) -> bool:
        """Check if the path is restricted.

        These paths are excluded from deletion and editing by the user.
        """
        relative_path = path.relative_to(self.project_folder)

        # Skip hidden files and directories (any path component starting with '.')
        # as well as `__pycache__` folders
        if any(part.startswith(".") for part in relative_path.parts):
            return True

        if "__pycache__" in relative_path.parts:
            return True

        # exclude the project_folder / models folder
        if relative_path.parts[0] == DEFAULT_MODELS_PATH:
            return True

        return False

    def bot_file_paths(
        self,
    ) -> Generator[Path, None, None]:
        """Get the paths of all bot files."""
        for file in self.project_folder.glob("**/*"):
            # Skip directories
            if not file.is_file() or self.is_restricted_path(file):
                continue

            yield file

    def _get_bot_data_for_llm(self) -> Dict[str, Any]:
        """Get the current bot data for the LLM."""
        file_importer = self._create_importer()

        # only include data created by the user (or the builder llm)
        # avoid including to many defaults that are not customized
        domain = file_importer.get_user_domain()
        flows = file_importer.get_user_flows()

        return {
            "domain": domain.as_dict(should_clean_json=True),
            "flows": yaml_flows_io.get_flows_as_json(flows, should_clean_json=True),
        }

    def _path_for_flow(self, flow_id: str) -> str:
        """Get the path for a flow."""
        if flow_id.startswith("pattern_"):
            return f"data/patterns/{flow_id}.yml"
        else:
            return f"data/flows/{flow_id}.yml"

    def _update_bot_files_from_llm_response(self, project_data: Dict[str, Any]) -> None:
        """Update the bot files with generated data by writing to disk."""
        files = {"domain.yml": dump_obj_as_yaml_to_string(project_data["domain"])}
        # split up flows into one file per flow in the /flows folder
        for flow_id, flow_data in project_data["flows"].get("flows", {}).items():
            flow_file_path = self._path_for_flow(flow_id)
            single_flow_file_data = {"flows": {flow_id: flow_data}}
            files[flow_file_path] = dump_obj_as_yaml_to_string(single_flow_file_data)

        # removes any other flows that the LLM didn't generate
        self._cleanup_flows()
        self.update_bot_files(files)

    def _cleanup_flows(self) -> None:
        """Cleanup the flows folder."""
        flows_folder = self.project_folder / "data" / "flows"
        if flows_folder.exists():
            shutil.rmtree(flows_folder)
        flows_folder.mkdir(parents=True, exist_ok=True)

    def update_bot_files(self, files: Dict[str, Optional[str]]) -> None:
        """Update bot files with new content by writing to disk."""
        for filename, content in files.items():
            file_path = Path(subpath(str(self.project_folder), filename))
            # Disallow updates inside .rasa project metadata directory
            if any(
                part.startswith(".")
                for part in file_path.relative_to(self.project_folder).parts
            ):
                # silently ignore hidden paths
                continue
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(content, encoding="utf-8")

    def ensure_all_files_are_writable(self, files: Dict[str, Optional[str]]) -> None:
        """Ensure all files are writable."""
        for filename, content in files.items():
            file_path = Path(subpath(str(self.project_folder), filename))
            if self.is_restricted_path(file_path):
                raise InvalidPathException(
                    f"This file or folder is restricted from editing: {file_path}"
                )

    def replace_all_bot_files(self, files: Dict[str, Optional[str]]) -> None:
        """Replace all bot files with new content, deleting files not in the request.

        Files/folders starting with .rasa/ or models/ are excluded from deletion.

        Args:
            files: Dictionary mapping file names to their content
        """
        self.ensure_all_files_are_writable(files)
        # Collect all existing files - any files not in the new `files` dict will be
        # deleted from this set
        existing_files = set(path.as_posix() for path in self.bot_file_paths())

        # Write all new files
        for filename, content in files.items():
            if content is None:
                continue

            file_path = Path(subpath(str(self.project_folder), filename))
            file_path.parent.mkdir(parents=True, exist_ok=True)

            try:
                file_path.write_text(content, encoding="utf-8")
            except Exception as e:
                # Log write failure and avoid deleting an existing file by mistake
                capture_exception_with_context(
                    e,
                    "project_generator.replace_all_bot_files.write_error",
                    extra={"file_path": file_path},
                )
                if file_path.as_posix() in existing_files:
                    # Keep the original file if it already existed
                    existing_files.discard(file_path.as_posix())
                continue

            # Remove from deletion set since this file is in the new set of files
            existing_files.discard(file_path.as_posix())

        # Delete files that weren't in the request
        for file_to_delete in existing_files:
            file_path = Path(file_to_delete)
            try:
                file_path.unlink()
            except Exception as e:
                capture_exception_with_context(
                    e,
                    "project_generator.replace_all_bot_files.delete_error",
                    extra={"file_path": file_path},
                )

        # Clean up empty directories (except excluded ones)
        self._cleanup_empty_directories()

    def _cleanup_empty_directories(self) -> None:
        """Remove empty directories from the project folder.

        Excludes hidden files and directories, and models/ from cleanup.
        """
        # Walk directories in reverse order (deepest first)
        for dirpath, dirnames, filenames in os.walk(self.project_folder, topdown=False):
            # Skip if this is the project root
            if dirpath == str(self.project_folder):
                continue

            if self.is_restricted_path(Path(dirpath)):
                continue

            relative_path = Path(dirpath).relative_to(self.project_folder)

            try:
                # Only remove if directory is empty
                if not os.listdir(dirpath):
                    os.rmdir(dirpath)
            except Exception as e:
                capture_exception_with_context(
                    e,
                    "project_generator.cleanup_empty_directories.error",
                    extra={"directory": relative_path.as_posix()},
                )

    def cleanup(self) -> None:
        """Cleanup the project folder."""
        # remove all the files and folders in the project folder resulting
        # in an empty folder
        for filename in os.listdir(self.project_folder):
            file_path = os.path.join(self.project_folder, filename)
            try:
                if filename == "lost+found":
                    continue
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                structlogger.error(
                    "project_generator.cleanup_error",
                    error=str(e),
                    file_path=file_path,
                )



import argparse
import logging
import os
from pathlib import Path
from typing import List, Optional, Text, Union

import rasa.cli.train as train
import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.utils.common
from rasa import model
from rasa.cli import SubParsersAction
from rasa.cli.arguments import interactive as arguments
from rasa.cli.validation.config_path_validation import get_validated_path
from rasa.core.config.configuration import Configuration
from rasa.core.constants import DEFAULT_SUB_AGENTS
from rasa.engine.storage.local_model_storage import LocalModelStorage
from rasa.shared.constants import (
    ASSISTANT_ID_DEFAULT_VALUE,
    ASSISTANT_ID_KEY,
    DEFAULT_ENDPOINTS_PATH,
    DEFAULT_MODELS_PATH,
)
from rasa.shared.data import TrainingType
from rasa.shared.importers.importer import TrainingDataImporter

logger = logging.getLogger(__name__)


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all interactive cli parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    interactive_parser = subparsers.add_parser(
        "interactive",
        conflict_handler="resolve",
        parents=parents,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Starts an interactive learning session to create new training data for a "
        "Rasa model by chatting.",
    )
    interactive_parser.set_defaults(func=interactive, core_only=False)

    interactive_subparsers = interactive_parser.add_subparsers()
    interactive_core_parser = interactive_subparsers.add_parser(
        "core",
        conflict_handler="resolve",
        parents=parents,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Starts an interactive learning session model to create new training data "
        "for a Rasa Core model by chatting. Uses the 'RegexMessageHandler', i.e. "
        "`/<intent>` input format.",
    )
    interactive_core_parser.set_defaults(func=interactive, core_only=True)

    arguments.set_interactive_arguments(interactive_parser)
    arguments.set_interactive_core_arguments(interactive_core_parser)


def interactive(args: argparse.Namespace) -> None:
    _set_not_required_args(args)
    file_importer = TrainingDataImporter.load_from_config(
        args.config, args.domain, args.data if not args.core_only else [args.stories]
    )

    if args.model is None:
        story_graph = file_importer.get_stories()
        if not story_graph or story_graph.is_empty():
            rasa.shared.utils.cli.print_error_and_exit(
                "Could not run interactive learning without either core "
                "data or a model containing core data."
            )

        zipped_model: Optional[Union[Text, Path]] = (
            train.run_core_training(args)
            if args.core_only
            else train.run_training(args)
        )
        if not zipped_model:
            rasa.shared.utils.cli.print_error_and_exit(
                "Could not train an initial model. Either pass paths "
                "to the relevant training files (`--data`, `--config`, `--domain`), "
                "or use 'rasa train' to train a model."
            )
    else:
        validate_assistant_id_key_in_config(file_importer)

        zipped_model = get_provided_model(args.model)
        if not (zipped_model and os.path.exists(zipped_model)):
            rasa.shared.utils.cli.print_error_and_exit(
                f"Interactive learning process cannot be started as no "
                f"initial model was found at path '{args.model}'.  "
                f"Use 'rasa train' to train a model."
            )
        if not args.skip_visualization:
            logger.info(f"Loading visualization data from {args.data}.")

    perform_interactive_learning(args, zipped_model, file_importer)


def _set_not_required_args(args: argparse.Namespace) -> None:
    args.fixed_model_name = None
    args.store_uncompressed = False
    args.dry_run = False
    args.skip_validation = True
    args.fail_on_validation_warnings = False
    args.validation_max_history = None
    args.remote_storage = None
    args.keep_local_model_copy = False
    args.sub_agents = DEFAULT_SUB_AGENTS  # Default value for sub-agents


def perform_interactive_learning(
    args: argparse.Namespace,
    zipped_model: Union[Text, "Path"],
    file_importer: TrainingDataImporter,
) -> None:
    """Performs interactive learning.

    Args:
        args: Namespace arguments.
        zipped_model: Path to zipped model.
        file_importer: File importer which provides the training data and model config.
    """
    from rasa.core.train import do_interactive_learning

    args.model = str(zipped_model)

    metadata = LocalModelStorage.metadata_from_archive(zipped_model)
    if metadata.training_type == TrainingType.NLU:
        rasa.shared.utils.cli.print_error_and_exit(
            "Can not run interactive learning on an NLU-only model."
        )

    args.endpoints = get_validated_path(
        args.endpoints, "endpoints", DEFAULT_ENDPOINTS_PATH, True
    )

    Configuration.initialise_endpoints(endpoints_path=Path(args.endpoints))

    do_interactive_learning(args, file_importer)


def get_provided_model(arg_model: Text) -> Optional[Union[Text, Path]]:
    """Checks model path input and selects model from it."""
    model_path = get_validated_path(arg_model, "model", DEFAULT_MODELS_PATH)

    return (
        model.get_latest_model(model_path) if os.path.isdir(model_path) else model_path
    )


def validate_assistant_id_key_in_config(file_importer: TrainingDataImporter) -> None:
    """Verifies that config contains a unique value for assistant identifier."""
    config_data = file_importer.get_config()
    assistant_id = config_data.get(ASSISTANT_ID_KEY)

    if assistant_id is None or assistant_id == ASSISTANT_ID_DEFAULT_VALUE:
        rasa.shared.utils.cli.print_error_and_exit(
            f"The '{ASSISTANT_ID_KEY}' key in the config file is either missing or "
            f"is set to the default value. Please replace the placeholder default "
            f"value and re-train the model."
        )
    return None



import os
from typing import TYPE_CHECKING, Any, Dict, Optional

from rasa.exceptions import HealthCheckError

if TYPE_CHECKING:
    pass
from rasa.shared.constants import (
    LLM_API_HEALTH_CHECK_DEFAULT_VALUE,
    LLM_API_HEALTH_CHECK_ENV_VAR,
    MODELS_CONFIG_KEY,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.providers.embedding.embedding_client import EmbeddingClient
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.utils.llm import embedder_factory, llm_factory, structlogger


def try_instantiate_llm_client(
    custom_llm_config: Optional[Dict],
    default_llm_config: Optional[Dict],
    log_source_function: str,
    log_source_component: str,
) -> LLMClient:
    """Validate llm configuration."""
    try:
        return llm_factory(custom_llm_config, default_llm_config)
    except (ProviderClientValidationError, ValueError) as e:
        raise HealthCheckError(
            code=f"{log_source_function}.llm_instantiation_failed",
            event_info=(
                f"Unable to create the LLM client for component - "
                f"{log_source_component}. "
                f"Please make sure you specified the required environment variables "
                f"and configuration keys. "
            ),
            error=e,
        )


def try_instantiate_embedder(
    custom_embeddings_config: Optional[Dict],
    default_embeddings_config: Optional[Dict],
    log_source_function: str,
    log_source_component: str,
) -> EmbeddingClient:
    """Validate embeddings configuration."""
    try:
        return embedder_factory(custom_embeddings_config, default_embeddings_config)
    except (ProviderClientValidationError, ValueError) as e:
        raise HealthCheckError(
            code=f"{log_source_function}.embedder_instantiation_failed",
            event_info=(
                f"Unable to create the Embedding client for component - "
                f"{log_source_component}. Please make sure you specified the required "
                f"environment variables and configuration keys."
            ),
            error=e,
        )


def perform_llm_health_check(
    custom_config: Optional[Dict[str, Any]],
    default_config: Dict[str, Any],
    log_source_function: str,
    log_source_component: str,
) -> None:
    """Try to instantiate the LLM Client to validate the provided config.

    If the LLM_API_HEALTH_CHECK environment variable is true, perform a test call
    to the LLM API. If config contains multiple models, perform a test call for each
    model in the model group.

    This method supports both single model configurations and model group configurations
    (configs that have the `models` key).
    """
    # Instantiate the LLM client or Router LLM client to validate the provided config.
    llm_client = try_instantiate_llm_client(
        custom_config, default_config, log_source_function, log_source_component
    )

    if is_api_health_check_enabled():
        if (
            custom_config
            and MODELS_CONFIG_KEY in custom_config
            and len(custom_config[MODELS_CONFIG_KEY]) > 1
        ):
            # If the config uses a router, instantiate the LLM client for each model
            # in the model group. This is required to perform a test api call for each
            # model in the group.
            # Note: The Router LLM client is not used here as we need to perform a test
            # api call and not load balance the requests.
            for model_config in custom_config[MODELS_CONFIG_KEY]:
                llm_client = try_instantiate_llm_client(
                    model_config,
                    default_config,
                    log_source_function,
                    log_source_component,
                )
                send_test_llm_api_request(
                    llm_client, log_source_function, log_source_component
                )
        else:
            # Make a test api call to perform a health check for the LLM client.
            # LLM config from config file and model group config from endpoint config
            # without router are handled here.
            send_test_llm_api_request(
                llm_client,
                log_source_function,
                log_source_component,
            )
    else:
        structlogger.warning(
            f"{log_source_function}.perform_llm_health_check.disabled",
            event_info=(
                f"The {LLM_API_HEALTH_CHECK_ENV_VAR} environment variable is set "
                f"to false, which will disable LLM health check. "
                f"It is recommended to set this variable to true in production "
                f"environments."
            ),
        )
        return None


def perform_embeddings_health_check(
    custom_config: Optional[Dict[str, Any]],
    default_config: Dict[str, Any],
    log_source_function: str,
    log_source_component: str,
) -> None:
    """Try to instantiate the Embedder to validate the provided config.

    If the LLM_API_HEALTH_CHECK environment variable is true, perform a test call
    to the Embeddings API. If config contains multiple models, perform a test call for
    each model in the model group.

    This method supports both single model configurations and model group configurations
    (configs that have the `models` key).
    """
    # Instantiate the Embedder client or the Embedder Router client to validate the
    # provided config. Deprecation warnings and errors are logged here.
    embedder = try_instantiate_embedder(
        custom_config, default_config, log_source_function, log_source_component
    )

    if is_api_health_check_enabled():
        if (
            custom_config
            and MODELS_CONFIG_KEY in custom_config
            and len(custom_config[MODELS_CONFIG_KEY]) > 1
        ):
            # If the config uses a router, instantiate the Embedder client for each
            # model in the model group. This is required to perform a test api call
            # for every model in the group.
            # Note: The Router Embedding client is not used here as we need to perform
            # a test API call and not load balance the requests.
            for model_config in custom_config[MODELS_CONFIG_KEY]:
                embedder = try_instantiate_embedder(
                    model_config,
                    default_config,
                    log_source_function,
                    log_source_component,
                )
                send_test_embeddings_api_request(
                    embedder, log_source_function, log_source_component
                )
        else:
            # Make a test api call to perform a health check for the Embedding client.
            # Embeddings config from config file and model group config from endpoint
            # config without router are handled here.
            send_test_embeddings_api_request(
                embedder, log_source_function, log_source_component
            )
    else:
        structlogger.warning(
            f"{log_source_function}" f".perform_embeddings_health_check.disabled",
            event_info=(
                f"The {LLM_API_HEALTH_CHECK_ENV_VAR} environment variable is set "
                f"to false, which will disable embeddings API health check. "
                f"It is recommended to set this variable to true in production "
                f"environments."
            ),
        )
        return None


def send_test_llm_api_request(
    llm_client: LLMClient, log_source_function: str, log_source_component: str
) -> None:
    """Sends a test request to the LLM API to perform a health check.

    Raises:
        Exception: If the API call fails.
    """
    structlogger.info(
        f"{log_source_function}.send_test_llm_api_request",
        event_info=(
            f"Sending a test LLM API request for the component - "
            f"{log_source_component}."
        ),
        config=llm_client.config,
    )
    try:
        llm_client.completion("hello")
    except Exception as e:
        raise HealthCheckError(
            code=f"{log_source_function}.send_test_llm_api_request_failed",
            event_info=(
                f"Test call to the LLM API failed for component - "
                f"{log_source_component}."
            ),
            config=llm_client.config,
            error=e,
        )


def send_test_embeddings_api_request(
    embedder: EmbeddingClient, log_source_function: str, log_source_component: str
) -> None:
    """Sends a test request to the Embeddings API to perform a health check.

    Raises:
        Exception: If the API call fails.
    """
    structlogger.info(
        f"{log_source_function}.send_test_embeddings_api_request",
        event_info=(
            f"Sending a test Embeddings API request for the component - "
            f"{log_source_component}."
        ),
        config=embedder.config,
    )
    try:
        embedder.embed(["hello"])
    except Exception as e:
        raise HealthCheckError(
            code=f"{log_source_function}.send_test_llm_api_request_failed",
            event_info=(
                f"Test call to the Embeddings API failed for component - "
                f"{log_source_component}."
            ),
            config=embedder.config,
            error=e,
        )


def is_api_health_check_enabled() -> bool:
    """Determines whether the API health check is enabled.

    Returns:
        bool: True if the API health check is enabled, False otherwise.
    """
    return (
        os.getenv(
            LLM_API_HEALTH_CHECK_ENV_VAR, LLM_API_HEALTH_CHECK_DEFAULT_VALUE
        ).lower()
        == "true"
    )



import importlib
import json
import os
import sys
from pathlib import Path
from types import FrameType
from typing import TYPE_CHECKING, Any, Dict, List, Text

import structlog

import rasa.shared.utils.io

if TYPE_CHECKING:
    from questionary import Question

structlogger = structlog.get_logger()

FREE_TEXT_INPUT_PROMPT = "Type out your own message..."


def parse_last_positional_argument_as_model_path() -> None:
    """Fixes the parsing of a potential positional model path argument."""
    if (
        len(sys.argv) >= 2
        # support relevant commands ...
        and sys.argv[1] in ["run", "shell", "interactive"]
        # but avoid interpreting subparser commands as model paths
        and sys.argv[1:] != ["run", "actions"]
        and not sys.argv[-2].startswith("-")
        and os.path.exists(sys.argv[-1])
    ):
        sys.argv.append(sys.argv[-1])
        sys.argv[-2] = "--model"


def button_to_string(button: Dict[Text, Any], idx: int = 0) -> Text:
    """Create a string representation of a button."""
    title = button.pop("title", "")

    if "payload" in button:
        payload = " ({})".format(button.pop("payload"))
    else:
        payload = ""

    # if there are any additional attributes, we append them to the output
    if button:
        details = " - {}".format(json.dumps(button, sort_keys=True))
    else:
        details = ""

    button_string = "{idx}: {title}{payload}{details}".format(
        idx=idx + 1, title=title, payload=payload, details=details
    )

    return button_string


def element_to_string(element: Dict[Text, Any], idx: int = 0) -> Text:
    """Create a string representation of an element."""
    title = element.pop("title", "")

    element_string = "{idx}: {title} - {element}".format(
        idx=idx + 1, title=title, element=json.dumps(element, sort_keys=True)
    )

    return element_string


def button_choices_from_message_data(
    message: Dict[Text, Any], allow_free_text_input: bool = True
) -> List[Text]:
    """Return list of choices to present to the user.

    If allow_free_text_input is True, an additional option is added
    at the end along with the response buttons that allows the user
    to type in free text.
    """
    choices = [
        button_to_string(button, idx)
        for idx, button in enumerate(message.get("buttons"))
    ]
    if allow_free_text_input:
        choices.append(FREE_TEXT_INPUT_PROMPT)
    return choices


async def payload_from_button_question(button_question: "Question") -> Text:
    """Prompt user with a button question and returns the nlu payload."""
    response = await button_question.ask_async()
    if response != FREE_TEXT_INPUT_PROMPT:
        # Find the button payload wrapped in (/payload)
        # The payload always starts with "/" so we look for "(/"
        payload_start = response.find("(/")
        if payload_start != -1:
            # Extract content between the outer parens: (/payload) → /payload
            payload_end = response.rfind(")")
            response = response[payload_start + 1 : payload_end]
    # Strip whitespace for clean output (handles free text with trailing newlines)
    return response.strip() if response else response


def signal_handler(_: int, __: FrameType) -> None:
    """Kills Rasa when OS signal is received."""
    print("Goodbye 👋")
    sys.exit(0)


def warn_if_rasa_plus_package_installed() -> None:
    """Issue a user warning in case the `rasa_plus` package is installed."""
    rasa_plus_package = "rasa_plus"
    if importlib.util.find_spec(rasa_plus_package) is not None:
        rasa.shared.utils.io.raise_warning(
            f"{rasa_plus_package} python package is no longer necessary "
            f"for using Rasa Pro. Please uninstall it.",
            UserWarning,
        )


def check_if_studio_command() -> bool:
    """Checks if the command is a Rasa Studio command."""
    return len(sys.argv) >= 2 and sys.argv[1] == "studio"


def get_e2e_results_file_name(
    results_output_path: Path,
    result_type: str,
) -> str:
    """Returns the name of the e2e results file."""
    if results_output_path.is_dir():
        file_name = results_output_path / f"e2e_results_{result_type}.yml"
    else:
        parent = results_output_path.parent
        stem = results_output_path.stem
        file_name = parent / f"{stem}_{result_type}.yml"

    return str(file_name)


def is_skip_validation_flag_set() -> bool:
    """Checks if the skip validation flag is set."""
    return "--skip-validation" in sys.argv



from typing import Any, Dict, Optional

import structlog
from sanic import Sanic

from rasa.builder import config
from rasa.builder.copilot.constants import (
    PROMPT_TO_BOT_KEY,
)
from rasa.builder.copilot.copilot_templated_message_provider import (
    load_copilot_handler_default_responses,
    load_copilot_welcome_messages,
)
from rasa.builder.copilot.models import (
    CopilotContext,
    FileContent,
    InternalCopilotRequestChatMessage,
    LogContent,
    ResponseCategory,
    TrainingErrorLog,
)
from rasa.builder.exceptions import (
    LLMGenerationError,
    ProjectGenerationError,
    TrainingError,
    ValidationError,
)
from rasa.builder.job_manager import JobInfo, job_manager
from rasa.builder.llm_service import llm_service
from rasa.builder.models import (
    JobStatus,
    JobStatusEvent,
)
from rasa.builder.project_generator import ProjectGenerator
from rasa.builder.training_service import (
    train_and_load_agent,
    try_load_existing_agent,
    update_agent,
)
from rasa.builder.validation_service import validate_project
from rasa.cli.scaffold import ProjectTemplateName

structlogger = structlog.get_logger()


async def push_job_status_event(
    job: JobInfo,
    status: JobStatus,
    message: Optional[str] = None,
    payload: Optional[Dict[str, Any]] = None,
) -> None:
    event = JobStatusEvent.from_status(
        status=status.value, message=message, payload=payload
    )
    job.status = status.value
    await job.put(event)


async def run_prompt_to_bot_job(
    app: Any,
    job: JobInfo,
    prompt: str,
) -> None:
    """Run the prompt-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        prompt: The natural language prompt for bot generation.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    await push_job_status_event(job, JobStatus.received)

    try:
        # 1. Generating
        await push_job_status_event(job, JobStatus.generating)
        bot_files = await project_generator.generate_project_with_retries(
            prompt,
            template=ProjectTemplateName.BASIC,
        )
        await push_job_status_event(job, JobStatus.generation_success)

        # 2. Training
        await push_job_status_event(job, JobStatus.training)
        agent = await train_and_load_agent(project_generator.get_training_input())
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        # 3. Create copilot welcome message job
        copilot_welcome_job = job_manager.create_job()
        app.add_task(run_copilot_welcome_message_job(app, copilot_welcome_job))

        structlogger.info(
            "bot_builder_service.prompt_to_bot.success",
            files_generated=list(bot_files.keys()),
            copilot_welcome_job_id=copilot_welcome_job.id,
        )
        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={"copilot_welcome_job_id": copilot_welcome_job.id},
        )
        job_manager.mark_done(job)

    except TrainingError as exc:
        structlogger.debug(
            "prompt_to_bot_job.training_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "prompt_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )

        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except (ProjectGenerationError, LLMGenerationError) as exc:
        structlogger.debug(
            "prompt_to_bot_job.generation_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "prompt_to_bot_job.unexpected_error", job_id=job.id, error=str(exc)
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_template_to_bot_job(
    app: "Sanic",
    job: JobInfo,
    template_name: ProjectTemplateName,
) -> None:
    """Run the template-to-bot job in the background.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        template_name: The name of the template to use for bot generation.
    """
    project_generator: ProjectGenerator = app.ctx.project_generator

    await push_job_status_event(job, JobStatus.received)

    try:
        # 1) Generating
        await push_job_status_event(job, JobStatus.generating)
        await project_generator.init_from_template(template_name)
        bot_files = project_generator.get_bot_files()
        await push_job_status_event(job, JobStatus.generation_success)

        # 2) Training
        await push_job_status_event(job, JobStatus.training)
        agent = await try_load_existing_agent(project_generator.project_folder)
        if agent is None:
            agent = await train_and_load_agent(project_generator.get_training_input())
        else:
            structlogger.info(
                "bot_builder_service.template_to_bot.agent_loaded_from_cache",
            )
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        # 3) Create copilot welcome message job
        copilot_welcome_job = job_manager.create_job()
        app.add_task(
            run_copilot_welcome_message_job(app, copilot_welcome_job, template_name)
        )

        # 4) Done - include welcome job ID in payload
        structlogger.info(
            "bot_builder_service.template_to_bot.success",
            files_generated=list(bot_files.keys()),
            copilot_welcome_job_id=copilot_welcome_job.id,
        )
        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={"copilot_welcome_job_id": copilot_welcome_job.id},
        )
        job_manager.mark_done(job)

    except TrainingError as exc:
        structlogger.debug(
            "template_to_bot_job.training_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.train_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except ValidationError as exc:
        # Log levels to include in the error message
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")

        structlogger.debug(
            "template_to_bot_job.validation_error",
            job_id=job.id,
            error=str(exc),
            all_validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        await push_job_status_event(
            job, JobStatus.validation_error, message=error_message
        )
        job_manager.mark_done(job, error=error_message)

    except ProjectGenerationError as exc:
        structlogger.debug(
            "template_to_bot_job.generation_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.generation_error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))

    except Exception as exc:
        # Capture full traceback
        structlogger.exception(
            "template_to_bot_job.unexpected_error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_replace_all_files_job(
    app: "Sanic",
    job: JobInfo,
    bot_files: Dict[str, Any],
) -> None:
    """Run the replace-all-files job in the background.

    This replaces all bot files with the provided files and deletes any files
    not included in the request (excluding .rasa/ and models/ directories).

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        bot_files: Dictionary of file names to content for replacement.
    """
    project_generator = app.ctx.project_generator
    await push_job_status_event(job, JobStatus.received)

    try:
        project_generator.replace_all_bot_files(bot_files)

        # Validating
        await push_job_status_event(job, JobStatus.validating)
        training_input = project_generator.get_training_input()
        validation_error = await validate_project(training_input.importer)
        if validation_error:
            raise ValidationError(validation_error)
        await push_job_status_event(job, JobStatus.validation_success)

        # Training
        await push_job_status_event(job, JobStatus.training)
        agent = await train_and_load_agent(training_input)
        update_agent(agent, app)
        await push_job_status_event(job, JobStatus.train_success)

        # Send final done event with copilot training success response job ID
        copilot_training_success_job = job_manager.create_job()
        app.add_task(
            run_copilot_training_success_job(app, copilot_training_success_job)
        )

        await push_job_status_event(
            job=job,
            status=JobStatus.done,
            payload={
                "copilot_training_success_job_id": copilot_training_success_job.id
            },
        )
        job_manager.mark_done(job)

    except ValidationError as exc:
        log_levels = ["error"]
        if config.VALIDATION_FAIL_ON_WARNINGS:
            log_levels.append("warning")
        structlogger.debug(
            "replace_all_files_job.validation_error",
            job_id=job.id,
            error=str(exc),
            validation_logs=exc.validation_logs,
            included_log_levels=log_levels,
        )
        error_message = exc.get_error_message_with_logs(log_levels=log_levels)
        # Push error event and start copilot analysis job
        await push_error_and_start_copilot_analysis(
            app,
            job,
            JobStatus.validation_error,
            error_message,
            bot_files,
        )

        # After error mark job as done
        job_manager.mark_done(job, error=error_message)

    except TrainingError as exc:
        error_message = str(exc)
        structlogger.debug(
            "replace_all_files_job.train_error",
            job_id=job.id,
            error=error_message,
        )
        # Push error event and start copilot analysis job
        await push_error_and_start_copilot_analysis(
            app,
            job,
            JobStatus.train_error,
            error_message,
            bot_files,
        )

        # After error mark job as done
        job_manager.mark_done(job, error=error_message)

    except Exception as exc:
        # Capture full traceback for anything truly unexpected
        error_message = str(exc)
        structlogger.exception(
            "replace_all_files_job.unexpected_error",
            job_id=job.id,
            error=error_message,
        )

        # Push error event and start copilot analysis job
        await push_error_and_start_copilot_analysis(
            app,
            job,
            JobStatus.error,
            error_message,
            bot_files,
        )

        # After error mark job as done
        job_manager.mark_done(job, error=str(exc))


async def push_error_and_start_copilot_analysis(
    app: "Sanic",
    original_job: JobInfo,
    original_job_status: JobStatus,
    error_message: str,
    bot_files: Dict[str, Any],
) -> None:
    """Start a copilot analysis job and notify the client.

    Creates a copilot analysis job and sends the new job ID to the client. The new
    job runs in the background.

    Args:
        app: The Sanic application instance
        original_job: The original job that failed
        original_job_status: The status of the job that failed
        error_message: The error message to analyze
        bot_files: The bot files to include in analysis
    """
    # Create a copilot analysis job. Send the new job ID to the client and
    # run the Copilot Analysis job in the background.
    message = "Failed to train the assistant. Starting copilot analysis."

    copilot_job = job_manager.create_job()
    # Push the error status event for the original job
    await push_job_status_event(
        original_job,
        original_job_status,
        message=message,
        payload={"copilot_job_id": copilot_job.id},
    )
    # Run the copilot analysis job in the background
    app.add_task(
        run_copilot_training_error_analysis_job(
            app, copilot_job, error_message, bot_files
        )
    )
    structlogger.debug(
        f"update_files_job.{original_job_status.value}.copilot_analysis_start",
        event_info=message,
        job_id=original_job.id,
        error=error_message,
        copilot_job_id=copilot_job.id,
    )


async def run_copilot_training_error_analysis_job(
    app: "Sanic",
    job: JobInfo,
    training_error_message: str,
    bot_files: Dict[str, Any],
) -> None:
    """Run copilot training error analysis job."""
    await push_job_status_event(job, JobStatus.received)

    try:
        # Create message content blocks with log content and available files
        log_content_block = LogContent(
            type="log", content=training_error_message, context="training_error"
        )
        file_content_blocks = [
            FileContent(type="file", file_path=file_path, file_content=file_content)
            for file_path, file_content in bot_files.items()
        ]
        context = CopilotContext(
            tracker_context=None,  # No conversation context needed
            copilot_chat_history=[
                InternalCopilotRequestChatMessage(
                    role="internal_copilot_request",
                    content=[log_content_block, *file_content_blocks],
                    response_category=ResponseCategory.TRAINING_ERROR_LOG_ANALYSIS,
                )
            ],
        )

        # Generate copilot response
        copilot_client = llm_service.instantiate_copilot()
        (
            original_stream,
            generation_context,
        ) = await copilot_client.generate_response(context)

        copilot_response_handler = llm_service.instantiate_handler(
            config.COPILOT_HANDLER_ROLLING_BUFFER_SIZE
        )
        intercepted_stream = copilot_response_handler.handle_response(original_stream)

        # Stream the copilot response as job events
        async for token in intercepted_stream:
            # Send each token as a job event using the same format as /copilot endpoint
            await push_job_status_event(
                job, JobStatus.copilot_analyzing, payload=token.sse_data
            )

        # Send references (if any) as part of copilot_analyzing stream
        if generation_context.relevant_documents:
            reference_section = copilot_response_handler.extract_references(
                generation_context.relevant_documents
            )
            await push_job_status_event(
                job, JobStatus.copilot_analyzing, payload=reference_section.sse_data
            )

        # Send original error log as part of copilot_analyzing stream
        training_error_log = TrainingErrorLog(logs=[log_content_block])
        await push_job_status_event(
            job, JobStatus.copilot_analyzing, payload=training_error_log.sse_data
        )

        # Send success status
        await push_job_status_event(job, JobStatus.copilot_analysis_success)

        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

    except Exception as exc:
        structlogger.exception(
            "copilot_training_error_analysis_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(
            job, JobStatus.copilot_analysis_error, message=str(exc)
        )
        job_manager.mark_done(job, error=str(exc))


async def run_copilot_welcome_message_job(
    app: "Sanic",
    job: JobInfo,
    template_name: Optional[ProjectTemplateName] = None,
) -> None:
    """Run the welcome message job in the background.

    This job sends a welcome message to the user after successful bot creation.
    For template-based bots, it sends a predefined message.
    For prompt-based bots, it can be extended to stream generated messages.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
        template_name: The template name for template-based bots, None for prompt-based.
    """
    try:
        # Load welcome messages from YAML
        welcome_messages = load_copilot_welcome_messages()

        # Get the appropriate welcome message
        if template_name:
            welcome_message = welcome_messages.get(
                template_name.value,
                welcome_messages.get(PROMPT_TO_BOT_KEY),
            )
        else:
            welcome_message = welcome_messages.get(PROMPT_TO_BOT_KEY)

        # Send the welcome message as a single event
        await push_job_status_event(
            job,
            JobStatus.copilot_welcome_message,
            payload={
                "content": welcome_message,
                "response_category": "copilot",
                "completeness": "complete",
            },
        )

        # Mark job as done
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

        structlogger.info(
            "copilot_welcome_message_job.success",
            job_id=job.id,
            template=template_name.value if template_name else PROMPT_TO_BOT_KEY,
        )

    except Exception as exc:
        structlogger.exception(
            "welcome_message_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))


async def run_copilot_training_success_job(
    app: "Sanic",
    job: JobInfo,
) -> None:
    """Run the training success job in the background.

    This job sends a training success message to the user after successful bot training.

    Args:
        app: The Sanic application instance.
        job: The job information instance.
    """
    try:
        # Load copilot default messages from YAML
        internal_messages = load_copilot_handler_default_responses()

        # Get the appropriate training success message
        training_success_message = internal_messages.get("training_success_response")

        # Send the training success message
        await push_job_status_event(
            job,
            JobStatus.train_success_message,
            payload={
                "content": training_success_message,
                "response_category": "copilot",
                "completeness": "complete",
            },
        )

        # Send the training success category
        await push_job_status_event(
            job,
            JobStatus.train_success_message,
            payload={
                "response_category": "copilot_training_success",
                "completeness": "complete",
            },
        )

        # Mark job as done
        await push_job_status_event(job, JobStatus.done)
        job_manager.mark_done(job)

        structlogger.info(
            "copilot_training_success_job.success",
            job_id=job.id,
        )

    except Exception as exc:
        structlogger.exception(
            "copilot_training_success_job.error",
            job_id=job.id,
            error=str(exc),
        )
        await push_job_status_event(job, JobStatus.error, message=str(exc))
        job_manager.mark_done(job, error=str(exc))



from __future__ import annotations

from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Optional,
    Set,
    Text,
)

import structlog

from rasa.shared.exceptions import RasaException

if TYPE_CHECKING:
    from rasa.shared.core.flows.flow_step_links import FlowStepLinks

structlogger = structlog.get_logger()


def step_from_json(flow_id: Text, data: Dict[Text, Any]) -> FlowStep:
    """Create a specific FlowStep from serialized data.

    Args:
        flow_id: The id of the flow that contains the step.
        data: data for a specific FlowStep object in a serialized data format.

    Returns:
        An instance of a specific FlowStep class.
    """
    from rasa.shared.core.flows.steps import (
        ActionFlowStep,
        CallFlowStep,
        CollectInformationFlowStep,
        LinkFlowStep,
        NoOperationFlowStep,
        SetSlotsFlowStep,
    )

    if "action" in data:
        return ActionFlowStep.from_json(flow_id, data)
    if "collect" in data:
        return CollectInformationFlowStep.from_json(flow_id, data)
    if "link" in data:
        return LinkFlowStep.from_json(flow_id, data)
    if "call" in data:
        return CallFlowStep.from_json(flow_id, data)
    if "set_slots" in data:
        return SetSlotsFlowStep.from_json(flow_id, data)
    if "noop" in data:
        return NoOperationFlowStep.from_json(flow_id, data)

    required_properties = ["action", "collect", "link", "call", "set_slots", "noop"]
    raise RasaException(
        f"Failed to parse step from json. Unknown type for {data}. "
        f"At least one of the following properties is required: "
        f"{', '.join(required_properties)}"
    )


@dataclass
class FlowStep:
    """A single step in a flow."""

    custom_id: Optional[Text]
    """The id of the flow step."""
    idx: int
    """The index of the step in the flow."""
    description: Optional[Text]
    """The description of the flow step."""
    metadata: Dict[Text, Any]
    """Additional, unstructured information about this flow step."""
    next: FlowStepLinks
    """The next steps of the flow step."""
    flow_id: Text
    """The id of the flow that contains the step."""

    @classmethod
    def from_json(cls, flow_id: Text, data: Dict[Text, Any]) -> FlowStep:
        """Create a FlowStep object from data in a serialized format.

        Args:
            flow_id: The id of the flow that contains the step.
            data: The data for a FlowStep object in a serialized format.

        Returns:
            The FlowStep object.
        """
        from rasa.shared.core.flows.flow_step_links import FlowStepLinks
        from rasa.shared.core.flows.steps.constants import UNSET_FLOW_STEP_ID

        return FlowStep(
            # the idx is set once the flow, which contains this step, is created
            idx=UNSET_FLOW_STEP_ID,
            custom_id=data.get("id"),
            description=data.get("description"),
            metadata=data.get("metadata", {}),
            next=FlowStepLinks.from_json(flow_id, data.get("next", [])),
            flow_id=flow_id,
        )

    def does_allow_for_next_step(self) -> bool:
        """Most steps allow linking to the next step. But some don't."""
        return True

    def as_json(
        self, step_properties: Optional[Dict[Text, Any]] = None
    ) -> Dict[Text, Any]:
        """Serialize the FlowStep object.

        Returns:
            The FlowStep as serialized data.
        """
        data: Dict[Text, Any] = {"id": self.id}

        if step_properties:
            data.update(step_properties)

        if self.description:
            data["description"] = self.description
        if dumped_next := self.next.as_json():
            data["next"] = dumped_next
        if self.metadata:
            data["metadata"] = self.metadata
        return data

    def steps_in_tree(
        self, should_resolve_calls: bool = True
    ) -> Generator[FlowStep, None, None]:
        """Recursively generates the steps in the tree."""
        yield self
        yield from self.next.steps_in_tree(should_resolve_calls)

    @property
    def id(self) -> Text:
        """Returns the id of the flow step."""
        return self.custom_id or self.default_id

    @property
    def default_id(self) -> str:
        """Returns the default id of the flow step."""
        return f"{self.flow_id}_{self.idx}_{self.default_id_postfix}"

    @property
    def default_id_postfix(self) -> str:
        """Returns the default id postfix of the flow step."""
        return "step"

    @property
    def utterances(self) -> Set[str]:
        """Return all the utterances used in this step."""
        return set()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FlowStep):
            return False

        return (
            self.idx == other.idx
            and self.description == other.description
            and self.next == other.next
            and self.flow_id == other.flow_id
        )


@dataclass
class FlowStepWithFlowReference:
    step: FlowStep
    """The step."""
    flow_id: str
    """The id of the flow that contains the step."""
